import os
import logging
import copy
import re

from cerberus import Validator

import itertools
import riscv_config.utils as utils
from riscv_config.errors import ValidationError
from riscv_config.schemaValidator import schemaValidator
import riscv_config.constants as constants
from riscv_config.warl import warl_interpreter

logger = logging.getLogger(__name__)


def reset():
    '''Function to set defaults to reset val of misa  based on presence of ISA extensions.'''
    global inp_yaml
    global extensions
    extension_enc = list("00000000000000000000000000")
    value=inp_yaml['ISA']
    if "32" in value:
       xlen = 32
       ext = value[4:]
    elif "64" in value:
       xlen = 64
       ext = value[4:]
    elif "128" in value:
       xlen = 128
       ext = value[5:]
    for x in "ABCDEFHIJKLMNPQSTUVX":
            if (x in ext):
                extension_enc[25 - int(ord(x) - ord('A'))] = "1"
    extensions = int("".join(extension_enc), 2)
    ext_b=format(extensions, '#0{}b'.format(xlen+2))
    mxl='10'if xlen==64 else '01'
    ext_b = ext_b[:2] + str(mxl) + ext_b[4:]
    return int(ext_b, 2)
    
def resetsu():
    '''Function to set defaults to reset val of mstatus based on the xlen and S, U extensions'''
    global inp_yaml
    if 64 in inp_yaml['supported_xlen'] and 'S' not in inp_yaml['ISA'] and 'U' in inp_yaml['ISA']:
      return 8589934592
    elif 64 in inp_yaml['supported_xlen'] and 'U' in inp_yaml['ISA'] and 'S' in inp_yaml['ISA']:
      return 42949672960
    else:	
      return 0
def reset_vsstatus():
    '''Function to set defaults to reset val of mstatus based on the xlen and S, U extensions'''
    global inp_yaml
    if 64 in inp_yaml['supported_xlen'] and 'U' in inp_yaml['ISA']:
      return 8589934592
    else:	
      return 0

def uset():
    '''Function to set defaults based on presence of 'U' extension.'''
    global inp_yaml
    if 'U' in inp_yaml['ISA']:
        return {'implemented': True}
    else:
        return {'implemented': False}
   
def hset():
    '''Function to set defaults based on presence of 'U' extension.'''
    global inp_yaml
    if 'H' in inp_yaml['ISA']:
        return {'implemented': True}
    else:
        return {'implemented': False}


def sset():
    '''Function to set defaults based on presence of 'S' extension.'''
    global inp_yaml
    if 'S' in inp_yaml['ISA']:
        return {'implemented': True}
    else:
        return {'implemented': False}
        
def fsset():
    '''Function to set defaults based on presence of 'F' extension.'''
    global inp_yaml
    if 'F' in inp_yaml['ISA'] or 'S' in inp_yaml['ISA']:
        return {'implemented': True}
    else:
        return {'implemented': False}
        

def uregset():
    '''Function to set defaults based on presence of 'U' extension.'''
    global inp_yaml
    temp = {'rv32': {'accessible': False}, 'rv64': {'accessible': False}}
    if 'U' in inp_yaml['ISA']:
      if 32 in inp_yaml['supported_xlen']:
        temp['rv32']['accessible'] = True
      if 64 in inp_yaml['supported_xlen']:
        temp['rv64']['accessible'] = True
    return temp

def uregseth():
    '''Function to set defaults based on presence of 'U' extension.'''
    global inp_yaml
    temp = {'rv32': {'accessible': False}, 'rv64': {'accessible': False}}
    if 'U' in inp_yaml['ISA']:
      if 32 in inp_yaml['supported_xlen']:
        temp['rv32']['accessible'] = True
    return temp
  
def hregseth():
    '''Function to set defaults based on presence of 'H' extension.'''
    global inp_yaml
    temp = {'rv32': {'accessible': False}, 'rv64': {'accessible': False}}
    if 'H' in inp_yaml['ISA']:
      if 32 in inp_yaml['supported_xlen']:
        temp['rv32']['accessible'] = True
    return temp      
def sregset():
    '''Function to set defaults based on presence of 'S' extension.'''
    global inp_yaml
    temp = {'rv32': {'accessible': False}, 'rv64': {'accessible': False}}
    if 'S' in inp_yaml['ISA']:
      if 32 in inp_yaml['supported_xlen']:
        temp['rv32']['accessible'] = True
      if 64 in inp_yaml['supported_xlen']:
        temp['rv64']['accessible'] = True
    return temp

def nregset():
    '''Function to set defaults based on presence of 'N' extension.'''
    global inp_yaml
    temp = {'rv32': {'accessible': False}, 'rv64': {'accessible': False}}
    if 'N' in inp_yaml['ISA']:
      if 32 in inp_yaml['supported_xlen']:
        temp['rv32']['accessible'] = True
      if 64 in inp_yaml['supported_xlen']:
        temp['rv64']['accessible'] = True
    return temp

def hregset():
    '''Function to set defaults based on presence of 'H' extension.'''
    global inp_yaml
    temp = {'rv32': {'accessible': False}, 'rv64': {'accessible': False}}
    if 'H' in inp_yaml['ISA']:
      if 32 in inp_yaml['supported_xlen']:
        temp['rv32']['accessible'] = True
      if 64 in inp_yaml['supported_xlen']:
        temp['rv64']['accessible'] = True
    return temp
    
def sregseth():
    '''Function to set defaults based on presence of 'S' extension.'''
    global inp_yaml
    temp = {'rv32': {'accessible': False}, 'rv64': {'accessible': False}}
    if 'S' in inp_yaml['ISA']:
      if 32 in inp_yaml['supported_xlen']:
        temp['rv32']['accessible'] = True
    return temp


def nuset():
    '''Function to check and set defaults for all fields which are dependent on
        the presence of 'U' extension and 'N' extension.'''
    global inp_yaml
    if 'U' in inp_yaml['ISA'] and 'N' in inp_yaml['ISA']:
        return {'implemented': True}
    else:
        return {'implemented': False}


def twset():
    '''Function to check and set value for tw field in misa.'''
    global inp_yaml
    if 'S' not in inp_yaml['ISA'] and 'U' not in inp_yaml['ISA']:
        return {'implemented': False}
    else:
        return {'implemented': True}


def delegset():
    '''Function to set "implemented" value for mideleg regisrer.'''
    # return True
    global inp_yaml
    var = True
    if 'S' not in inp_yaml['ISA'] and 'N' not in inp_yaml['ISA']:
        var = False

    temp = {'rv32': {'accessible': False}, 'rv64': {'accessible': False}}
    if 32 in inp_yaml['supported_xlen']:
        temp['rv32']['accessible'] = True and var
    if 64 in inp_yaml['supported_xlen']:
        temp['rv64']['accessible'] = True and var
    return temp


def countset():
    global inp_yaml
    temp = {'rv32': {'accessible': False}, 'rv64': {'accessible': False}}
    if 'S' in inp_yaml['ISA'] or 'U' in inp_yaml["ISA"]:
        if 32 in inp_yaml['supported_xlen']:
            temp['rv32']['accessible'] = True
        if 64 in inp_yaml['supported_xlen']:
            temp['rv64']['accessible'] = True
    return temp


def regset():
    global inp_yaml
    temp = {'rv32': {'accessible': False}, 'rv64': {'accessible': False}}
    if 32 in inp_yaml['supported_xlen']:
        temp['rv32']['accessible'] = True
    if 64 in inp_yaml['supported_xlen']:
        temp['rv64']['accessible'] = True
    return temp
    
def pmpregset():
    global inp_yaml
    temp = {'rv32': {'accessible': False}, 'rv64': {'accessible': False}} 
    try:   
     if 32 in inp_yaml['supported_xlen'] and inp_yaml[temp]['rv32']['accessible'] :
        temp['rv32']['accessible'] = True
    except:
        temp['rv32']['accessible'] = False
    try: 
     if 64 in inp_yaml['supported_xlen'] and inp_yaml[temp]['rv64']['accessible']:
        temp['rv64']['accessible'] = True
    except:
        temp['rv64']['accessible'] = False
    return temp    

def pmpcounterhset():
    global inp_yaml
    temp = {'rv32': {'accessible': False}, 'rv64': {'accessible': False}}
    try: 
     if 32 in inp_yaml['supported_xlen'] and inp_yaml[temp]['rv32']['accessible'] :
        temp['rv32']['accessible'] = True
    except:
        temp['rv32']['accessible'] = False
    return temp
    
def counterhset():
    global inp_yaml
    temp = {'rv32': {'accessible': False}, 'rv64': {'accessible': False}}
    if 32 in inp_yaml['supported_xlen']:
        temp['rv32']['accessible'] = True
    return temp

def add_debug_setters(schema_yaml):
    '''Function to set the default setters for various fields in the debug schema'''
    regsetter = lambda doc: regset()
    
    schema_yaml['dcsr']['default_setter'] = regsetter
    return schema_yaml
    
def add_reset_setters(schema_yaml):
    '''Function to set the default setters for extension  subfields in the misa'''
    global inp_yaml
    global extensions
    xlen=inp_yaml['supported_xlen'][0]
    rvxlen='rv'+str(xlen)
    extensions=hex(int(format(reset(), '#0{}b'.format(xlen+2))[(xlen-24):(xlen+2)], 2))
    schema_yaml['misa']['schema'][rvxlen]['schema']['extensions']['schema']['type']['default']['warl']['legal'][0]=schema_yaml['misa']['schema'][rvxlen]['schema']['extensions']['schema']['type']['default']['warl']['legal'][0].replace('0x3FFFFFFF', extensions)
    return schema_yaml
    
def add_fflags_type_setters(schema_yaml):
    global inp_yaml
    xlen=inp_yaml['supported_xlen'][0]
    rvxlen='rv'+str(xlen)
    if 'F' not in inp_yaml['ISA']:
        schema_yaml['fcsr']['schema'][rvxlen]['schema']['frm']['schema']['type']['default'] = {'ro_constant': 0}
        schema_yaml['fcsr']['schema'][rvxlen]['schema']['fflags']['schema']['type']['default'] = {'ro_constant': 0}
    return schema_yaml
    
def add_def_setters(schema_yaml):
    '''Function to set the default setters for various fields in the schema'''
    regsetter = lambda doc: regset()
    resetsetter=lambda doc: reset()
    reset_susetter=lambda doc: resetsu()
    reset_vsssetter=lambda doc: reset_vsstatus()
    pmpregsetter = lambda doc: pmpregset()
    counthsetter = lambda doc: counterhset()
    pmpcounthsetter = lambda doc: pmpcounterhset()
    uregsetter = lambda doc: uregset()
    ureghsetter = lambda doc: uregseth()
    ssetter = lambda doc: sset()
    fssetter = lambda doc: fsset()
    sregsetter = lambda doc: sregset()
    nregsetter = lambda doc: nregset()
    hregsetter = lambda doc: hregset()
    hreghsetter = lambda doc: hregseth()
    sregsetterh = lambda doc: sregseth()
    nusetter = lambda doc: nuset()
    usetter = lambda doc: uset()
    hsetter = lambda doc: hset()
    twsetter = lambda doc: twset()
    delegsetter = lambda doc: delegset()

    schema_yaml['sstatus']['default_setter'] = sregsetter
    schema_yaml['sstatus']['schema']['rv32']['schema']['uie'][
        'default_setter'] = nusetter
    schema_yaml['sstatus']['schema']['rv64']['schema']['uie'][
        'default_setter'] = nusetter
    schema_yaml['sstatus']['schema']['rv32']['schema']['upie'][
        'default_setter'] = nusetter
    schema_yaml['sstatus']['schema']['rv64']['schema']['upie'][
        'default_setter'] = nusetter

    schema_yaml['sstatus']['schema']['rv64']['schema']['uxl'][
        'default_setter'] = usetter
    schema_yaml['sstatus']['schema']['rv32']['schema']['sie'][
        'default_setter'] = ssetter
    schema_yaml['sstatus']['schema']['rv64']['schema']['sie'][
        'default_setter'] = ssetter
    schema_yaml['sstatus']['schema']['rv32']['schema']['spie'][
        'default_setter'] = ssetter
    schema_yaml['sstatus']['schema']['rv64']['schema']['spie'][
        'default_setter'] = ssetter
    schema_yaml['sstatus']['schema']['rv32']['schema']['spp'][
        'default_setter'] = ssetter
    schema_yaml['sstatus']['schema']['rv64']['schema']['spp'][
        'default_setter'] = ssetter
    schema_yaml['sstatus']['schema']['rv32']['schema']['mxr'][
        'default_setter'] = ssetter
    schema_yaml['sstatus']['schema']['rv64']['schema']['mxr'][
        'default_setter'] = ssetter
    schema_yaml['sstatus']['schema']['rv32']['schema']['sum'][
        'default_setter'] = ssetter
    schema_yaml['sstatus']['schema']['rv64']['schema']['sum'][
        'default_setter'] = ssetter
    schema_yaml['sstatus']['schema']['rv32']['schema']['fs'][
        'default_setter'] = fssetter
    schema_yaml['sstatus']['schema']['rv64']['schema']['fs'][
        'default_setter'] = fssetter
    schema_yaml['sstatus']['schema']['rv32']['schema']['sd'][
        'default_setter'] = fssetter
    schema_yaml['sstatus']['schema']['rv64']['schema']['sd'][
        'default_setter'] = fssetter
    schema_yaml['sie']['default_setter'] = sregsetter
    schema_yaml['sie']['schema']['rv32']['schema']['ueie'][
        'default_setter'] = nusetter
    schema_yaml['sie']['schema']['rv64']['schema']['ueie'][
        'default_setter'] = nusetter
    schema_yaml['sie']['schema']['rv32']['schema']['utie'][
        'default_setter'] = nusetter
    schema_yaml['sie']['schema']['rv64']['schema']['utie'][
        'default_setter'] = nusetter
    schema_yaml['sie']['schema']['rv32']['schema']['usie'][
        'default_setter'] = nusetter
    schema_yaml['sie']['schema']['rv64']['schema']['usie'][
        'default_setter'] = nusetter
    schema_yaml['sie']['schema']['rv32']['schema']['seie'][
        'default_setter'] = ssetter
    schema_yaml['sie']['schema']['rv64']['schema']['seie'][
        'default_setter'] = ssetter
    schema_yaml['sie']['schema']['rv32']['schema']['stie'][
        'default_setter'] = ssetter
    schema_yaml['sie']['schema']['rv64']['schema']['stie'][
        'default_setter'] = ssetter
    schema_yaml['sie']['schema']['rv32']['schema']['ssie'][
        'default_setter'] = ssetter
    schema_yaml['sie']['schema']['rv64']['schema']['ssie'][
        'default_setter'] = ssetter
    schema_yaml['sip']['default_setter'] = sregsetter
    schema_yaml['sip']['schema']['rv32']['schema']['ueip'][
        'default_setter'] = nusetter
    schema_yaml['sip']['schema']['rv64']['schema']['ueip'][
        'default_setter'] = nusetter
    schema_yaml['sip']['schema']['rv32']['schema']['utip'][
        'default_setter'] = nusetter
    schema_yaml['sip']['schema']['rv64']['schema']['utip'][
        'default_setter'] = nusetter
    schema_yaml['sip']['schema']['rv32']['schema']['usip'][
        'default_setter'] = nusetter
    schema_yaml['sip']['schema']['rv64']['schema']['usip'][
        'default_setter'] = nusetter
    schema_yaml['sip']['schema']['rv32']['schema']['seip'][
        'default_setter'] = ssetter
    schema_yaml['sip']['schema']['rv64']['schema']['seip'][
        'default_setter'] = ssetter
    schema_yaml['sip']['schema']['rv32']['schema']['stip'][
        'default_setter'] = ssetter
    schema_yaml['sip']['schema']['rv64']['schema']['stip'][
        'default_setter'] = ssetter
    schema_yaml['sip']['schema']['rv32']['schema']['ssip'][
        'default_setter'] = ssetter
    schema_yaml['sip']['schema']['rv64']['schema']['ssip'][
        'default_setter'] = ssetter
    schema_yaml['stvec']['default_setter'] = sregsetter
    schema_yaml['stvec']['schema']['rv32']['schema']['base'][
        'default_setter'] = ssetter
    schema_yaml['stvec']['schema']['rv64']['schema']['base'][
        'default_setter'] = ssetter
    schema_yaml['stvec']['schema']['rv32']['schema']['mode'][
        'default_setter'] = ssetter
    schema_yaml['stvec']['schema']['rv64']['schema']['mode'][
        'default_setter'] = ssetter
        
    schema_yaml['sepc']['default_setter'] = sregsetter
    schema_yaml['stval']['default_setter'] = sregsetter
    schema_yaml['scause']['default_setter'] = sregsetter
    schema_yaml['scause']['schema']['rv32']['schema']['interrupt'][
        'default_setter'] = ssetter
    schema_yaml['scause']['schema']['rv64']['schema']['interrupt'][
        'default_setter'] = ssetter
    schema_yaml['scause']['schema']['rv32']['schema']['exception_code'][
        'default_setter'] = ssetter
    schema_yaml['scause']['schema']['rv64']['schema']['exception_code'][
        'default_setter'] = ssetter
    schema_yaml['satp']['default_setter'] = sregsetter
    schema_yaml['satp']['schema']['rv32']['schema']['ppn'][
        'default_setter'] = ssetter
    schema_yaml['satp']['schema']['rv64']['schema']['ppn'][
        'default_setter'] = ssetter
    schema_yaml['satp']['schema']['rv32']['schema']['asid'][
        'default_setter'] = ssetter
    schema_yaml['satp']['schema']['rv64']['schema']['asid'][
        'default_setter'] = ssetter
    schema_yaml['satp']['schema']['rv32']['schema']['mode'][
        'default_setter'] = ssetter
    schema_yaml['satp']['schema']['rv64']['schema']['mode'][
        'default_setter'] = ssetter
    schema_yaml['sscratch']['default_setter'] = sregsetter
    
    schema_yaml['ustatus']['default_setter'] = nregsetter
    schema_yaml['ustatus']['schema']['rv32']['schema']['uie'][
        'default_setter'] = nusetter
    schema_yaml['ustatus']['schema']['rv64']['schema']['uie'][
        'default_setter'] = nusetter
    schema_yaml['ustatus']['schema']['rv32']['schema']['upie'][
        'default_setter'] = nusetter
    schema_yaml['ustatus']['schema']['rv64']['schema']['upie'][
        'default_setter'] = nusetter

    schema_yaml['uie']['default_setter'] = nregsetter
    schema_yaml['uie']['schema']['rv32']['schema']['ueie'][
        'default_setter'] = nusetter
    schema_yaml['uie']['schema']['rv64']['schema']['ueie'][
        'default_setter'] = nusetter
    schema_yaml['uie']['schema']['rv32']['schema']['utie'][
        'default_setter'] = nusetter
    schema_yaml['uie']['schema']['rv64']['schema']['utie'][
        'default_setter'] = nusetter
    schema_yaml['uie']['schema']['rv32']['schema']['usie'][
        'default_setter'] = nusetter
    schema_yaml['uie']['schema']['rv64']['schema']['usie'][
        'default_setter'] = nusetter
    schema_yaml['uip']['default_setter'] = nregsetter
    schema_yaml['uip']['schema']['rv32']['schema']['ueip'][
        'default_setter'] = nusetter
    schema_yaml['uip']['schema']['rv64']['schema']['ueip'][
        'default_setter'] = nusetter
    schema_yaml['uip']['schema']['rv32']['schema']['utip'][
        'default_setter'] = nusetter
    schema_yaml['uip']['schema']['rv64']['schema']['utip'][
        'default_setter'] = nusetter
    schema_yaml['uip']['schema']['rv32']['schema']['usip'][
        'default_setter'] = nusetter
    schema_yaml['uip']['schema']['rv64']['schema']['usip'][
        'default_setter'] = nusetter
    schema_yaml['utvec']['default_setter'] = nregsetter
    schema_yaml['utvec']['schema']['rv32']['schema']['base'][
        'default_setter'] = nusetter
    schema_yaml['utvec']['schema']['rv64']['schema']['base'][
        'default_setter'] = nusetter
    schema_yaml['utvec']['schema']['rv32']['schema']['mode'][
        'default_setter'] = nusetter
    schema_yaml['utvec']['schema']['rv64']['schema']['mode'][
        'default_setter'] = nusetter
    schema_yaml['uepc']['default_setter'] = nregsetter
    schema_yaml['utval']['default_setter'] = nregsetter
    schema_yaml['ucause']['default_setter'] = nregsetter
    schema_yaml['ucause']['schema']['rv32']['schema']['interrupt'][
        'default_setter'] = nusetter
    schema_yaml['ucause']['schema']['rv64']['schema']['interrupt'][
        'default_setter'] = nusetter
    schema_yaml['ucause']['schema']['rv32']['schema']['exception_code'][
        'default_setter'] = nusetter
    schema_yaml['ucause']['schema']['rv64']['schema']['exception_code'][
        'default_setter'] = nusetter
    schema_yaml['uscratch']['default_setter'] = nregsetter

    schema_yaml['fcsr']['schema']['rv32']['schema']['frm'][
        'default_setter'] = usetter
    schema_yaml['fcsr']['schema']['rv64']['schema']['frm'][
        'default_setter'] = usetter
    schema_yaml['fcsr']['schema']['rv32']['schema']['fflags'][
        'default_setter'] = usetter
    schema_yaml['fcsr']['schema']['rv64']['schema']['fflags'][
        'default_setter'] = usetter

    schema_yaml['misa']['default_setter'] = regsetter
    schema_yaml['misa']['schema']['reset-val']['default_setter'] = resetsetter
    schema_yaml['mstatus']['default_setter'] = regsetter
    schema_yaml['mstatus']['schema']['reset-val']['default_setter']=reset_susetter
    schema_yaml['vsstatus']['schema']['reset-val']['default_setter']=reset_vsssetter
    schema_yaml['mstatush']['default_setter'] = counthsetter   
    schema_yaml['mvendorid']['default_setter'] = regsetter
    schema_yaml['mimpid']['default_setter'] = regsetter
    schema_yaml['marchid']['default_setter'] = regsetter
    schema_yaml['mhartid']['default_setter'] = regsetter
    schema_yaml['mtvec']['default_setter'] = regsetter
    schema_yaml['mip']['default_setter'] = regsetter
    schema_yaml['mie']['default_setter'] = regsetter
    schema_yaml['mscratch']['default_setter'] = regsetter
    schema_yaml['mepc']['default_setter'] = regsetter
    schema_yaml['mtval']['default_setter'] = regsetter
    schema_yaml['mtval2']['default_setter'] = hregsetter
    schema_yaml['mtinst']['default_setter'] = hregsetter
    schema_yaml['mcountinhibit']['default_setter'] = regsetter
    schema_yaml['mcycle']['default_setter'] = regsetter
    schema_yaml['minstret']['default_setter'] = regsetter
    schema_yaml['mcycleh']['default_setter'] = counthsetter
    schema_yaml['minstreth']['default_setter'] = counthsetter
    schema_yaml['pmpcfg0']['default_setter'] = pmpregsetter
    schema_yaml['pmpcfg1']['default_setter'] = pmpcounthsetter
    schema_yaml['pmpcfg2']['default_setter'] = pmpregsetter
    schema_yaml['pmpcfg3']['default_setter'] = pmpcounthsetter
    schema_yaml['pmpcfg4']['default_setter'] = pmpregsetter
    schema_yaml['pmpcfg5']['default_setter'] = pmpcounthsetter
    schema_yaml['pmpcfg6']['default_setter'] = pmpregsetter
    schema_yaml['pmpcfg7']['default_setter'] = pmpcounthsetter
    schema_yaml['pmpcfg8']['default_setter'] = pmpregsetter
    schema_yaml['pmpcfg9']['default_setter'] = pmpcounthsetter
    schema_yaml['pmpcfg10']['default_setter'] = pmpregsetter
    schema_yaml['pmpcfg11']['default_setter'] = pmpcounthsetter
    schema_yaml['pmpcfg12']['default_setter'] = pmpregsetter
    schema_yaml['pmpcfg13']['default_setter'] = pmpcounthsetter
    schema_yaml['pmpcfg14']['default_setter'] = pmpregsetter
    schema_yaml['pmpcfg15']['default_setter'] = pmpcounthsetter
    schema_yaml['pmpaddr0']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr1']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr2']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr3']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr4']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr5']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr6']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr7']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr8']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr9']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr10']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr11']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr12']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr13']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr14']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr15']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr16']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr17']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr18']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr19']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr20']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr21']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr22']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr23']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr24']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr25']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr26']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr27']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr28']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr29']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr30']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr31']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr32']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr33']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr34']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr35']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr36']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr37']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr38']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr39']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr40']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr41']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr42']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr43']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr44']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr45']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr46']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr47']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr48']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr49']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr50']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr51']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr52']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr53']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr54']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr55']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr56']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr57']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr58']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr59']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr60']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr61']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr62']['default_setter'] = pmpregsetter
    schema_yaml['pmpaddr63']['default_setter'] = pmpregsetter

    # event counters
    schema_yaml['mhpmevent3']['default_setter'] = regsetter
    schema_yaml['mhpmevent4']['default_setter'] = regsetter
    schema_yaml['mhpmevent5']['default_setter'] = regsetter
    schema_yaml['mhpmevent6']['default_setter'] = regsetter
    schema_yaml['mhpmevent7']['default_setter'] = regsetter
    schema_yaml['mhpmevent8']['default_setter'] = regsetter
    schema_yaml['mhpmevent9']['default_setter'] = regsetter
    schema_yaml['mhpmevent10']['default_setter'] = regsetter
    schema_yaml['mhpmevent11']['default_setter'] = regsetter
    schema_yaml['mhpmevent12']['default_setter'] = regsetter
    schema_yaml['mhpmevent13']['default_setter'] = regsetter
    schema_yaml['mhpmevent14']['default_setter'] = regsetter
    schema_yaml['mhpmevent15']['default_setter'] = regsetter
    schema_yaml['mhpmevent16']['default_setter'] = regsetter
    schema_yaml['mhpmevent17']['default_setter'] = regsetter
    schema_yaml['mhpmevent18']['default_setter'] = regsetter
    schema_yaml['mhpmevent19']['default_setter'] = regsetter
    schema_yaml['mhpmevent20']['default_setter'] = regsetter
    schema_yaml['mhpmevent21']['default_setter'] = regsetter
    schema_yaml['mhpmevent22']['default_setter'] = regsetter
    schema_yaml['mhpmevent23']['default_setter'] = regsetter
    schema_yaml['mhpmevent24']['default_setter'] = regsetter
    schema_yaml['mhpmevent25']['default_setter'] = regsetter
    schema_yaml['mhpmevent26']['default_setter'] = regsetter
    schema_yaml['mhpmevent27']['default_setter'] = regsetter
    schema_yaml['mhpmevent28']['default_setter'] = regsetter
    schema_yaml['mhpmevent29']['default_setter'] = regsetter
    schema_yaml['mhpmevent30']['default_setter'] = regsetter
    schema_yaml['mhpmevent31']['default_setter'] = regsetter

    schema_yaml['mhpmcounter3']['default_setter'] = regsetter
    schema_yaml['mhpmcounter4']['default_setter'] = regsetter
    schema_yaml['mhpmcounter5']['default_setter'] = regsetter
    schema_yaml['mhpmcounter6']['default_setter'] = regsetter
    schema_yaml['mhpmcounter7']['default_setter'] = regsetter
    schema_yaml['mhpmcounter8']['default_setter'] = regsetter
    schema_yaml['mhpmcounter9']['default_setter'] = regsetter
    schema_yaml['mhpmcounter10']['default_setter'] = regsetter
    schema_yaml['mhpmcounter11']['default_setter'] = regsetter
    schema_yaml['mhpmcounter12']['default_setter'] = regsetter
    schema_yaml['mhpmcounter13']['default_setter'] = regsetter
    schema_yaml['mhpmcounter14']['default_setter'] = regsetter
    schema_yaml['mhpmcounter15']['default_setter'] = regsetter
    schema_yaml['mhpmcounter16']['default_setter'] = regsetter
    schema_yaml['mhpmcounter17']['default_setter'] = regsetter
    schema_yaml['mhpmcounter18']['default_setter'] = regsetter
    schema_yaml['mhpmcounter19']['default_setter'] = regsetter
    schema_yaml['mhpmcounter20']['default_setter'] = regsetter
    schema_yaml['mhpmcounter21']['default_setter'] = regsetter
    schema_yaml['mhpmcounter22']['default_setter'] = regsetter
    schema_yaml['mhpmcounter23']['default_setter'] = regsetter
    schema_yaml['mhpmcounter24']['default_setter'] = regsetter
    schema_yaml['mhpmcounter25']['default_setter'] = regsetter
    schema_yaml['mhpmcounter26']['default_setter'] = regsetter
    schema_yaml['mhpmcounter27']['default_setter'] = regsetter
    schema_yaml['mhpmcounter28']['default_setter'] = regsetter
    schema_yaml['mhpmcounter29']['default_setter'] = regsetter
    schema_yaml['mhpmcounter30']['default_setter'] = regsetter
    schema_yaml['mhpmcounter31']['default_setter'] = regsetter

    schema_yaml['mhpmcounter3h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter4h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter5h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter6h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter7h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter8h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter9h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter10h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter11h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter12h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter13h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter14h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter15h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter16h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter17h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter18h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter19h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter20h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter21h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter22h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter23h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter24h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter25h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter26h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter27h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter28h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter29h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter30h']['default_setter'] = counthsetter
    schema_yaml['mhpmcounter31h']['default_setter'] = counthsetter
    schema_yaml['hpmcounter3']['default_setter'] = uregsetter
    schema_yaml['hpmcounter3h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter4']['default_setter'] = uregsetter
    schema_yaml['hpmcounter4h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter5']['default_setter'] = uregsetter
    schema_yaml['hpmcounter5h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter6']['default_setter'] = uregsetter
    schema_yaml['hpmcounter6h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter7']['default_setter'] = uregsetter
    schema_yaml['hpmcounter7h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter8']['default_setter'] = uregsetter
    schema_yaml['hpmcounter8h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter9']['default_setter'] = uregsetter
    schema_yaml['hpmcounter9h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter10']['default_setter'] = uregsetter
    schema_yaml['hpmcounter10h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter11']['default_setter'] = uregsetter
    schema_yaml['hpmcounter11h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter12']['default_setter'] = uregsetter
    schema_yaml['hpmcounter12h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter13']['default_setter'] = uregsetter
    schema_yaml['hpmcounter13h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter14']['default_setter'] = uregsetter
    schema_yaml['hpmcounter14h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter15']['default_setter'] = uregsetter
    schema_yaml['hpmcounter15h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter16']['default_setter'] = uregsetter
    schema_yaml['hpmcounter16h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter17']['default_setter'] = uregsetter
    schema_yaml['hpmcounter17h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter18']['default_setter'] = uregsetter
    schema_yaml['hpmcounter18h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter19']['default_setter'] = uregsetter
    schema_yaml['hpmcounter19h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter20']['default_setter'] = uregsetter
    schema_yaml['hpmcounter20h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter21']['default_setter'] = uregsetter
    schema_yaml['hpmcounter21h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter22']['default_setter'] = uregsetter
    schema_yaml['hpmcounter22h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter23']['default_setter'] = uregsetter
    schema_yaml['hpmcounter23h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter24']['default_setter'] = uregsetter
    schema_yaml['hpmcounter24h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter25']['default_setter'] = uregsetter
    schema_yaml['hpmcounter25h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter26']['default_setter'] = uregsetter
    schema_yaml['hpmcounter26h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter27']['default_setter'] = uregsetter
    schema_yaml['hpmcounter27h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter28']['default_setter'] = uregsetter
    schema_yaml['hpmcounter28h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter29']['default_setter'] = uregsetter
    schema_yaml['hpmcounter29h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter30']['default_setter'] = uregsetter
    schema_yaml['hpmcounter30h']['default_setter'] = ureghsetter
    schema_yaml['hpmcounter31']['default_setter'] = uregsetter
    schema_yaml['hpmcounter31h']['default_setter'] = ureghsetter

    schema_yaml['mcounteren']['default_setter'] = uregsetter
    schema_yaml['scounteren']['default_setter'] = uregsetter

    schema_yaml['mcause']['default_setter'] = regsetter
    schema_yaml['mstatus']['schema']['rv32']['schema']['uie'][
        'default_setter'] = nusetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['uie'][
        'default_setter'] = nusetter
    schema_yaml['mstatus']['schema']['rv32']['schema']['upie'][
        'default_setter'] = nusetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['upie'][
        'default_setter'] = nusetter

    schema_yaml['mstatus']['schema']['rv32']['schema']['mprv'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['mprv'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['uxl'][
        'default_setter'] = usetter
    schema_yaml['fflags']['default_setter'] = uregsetter
    schema_yaml['frm']['default_setter'] = uregsetter
    schema_yaml['fcsr']['default_setter'] = uregsetter
    schema_yaml['time']['default_setter'] = uregsetter
    schema_yaml['timeh']['default_setter'] = ureghsetter
    schema_yaml['cycle']['default_setter'] = uregsetter
    schema_yaml['cycleh']['default_setter'] = ureghsetter
    schema_yaml['instret']['default_setter'] = uregsetter
    schema_yaml['instreth']['default_setter'] = ureghsetter

    schema_yaml['mip']['schema']['rv32']['schema']['ueip'][
        'default_setter'] = nusetter
    schema_yaml['mip']['schema']['rv64']['schema']['ueip'][
        'default_setter'] = nusetter
    schema_yaml['mip']['schema']['rv32']['schema']['utip'][
        'default_setter'] = nusetter
    schema_yaml['mip']['schema']['rv64']['schema']['utip'][
        'default_setter'] = nusetter
    schema_yaml['mip']['schema']['rv32']['schema']['usip'][
        'default_setter'] = nusetter
    schema_yaml['mip']['schema']['rv64']['schema']['usip'][
        'default_setter'] = nusetter
    schema_yaml['mie']['schema']['rv32']['schema']['ueie'][
        'default_setter'] = nusetter
    schema_yaml['mie']['schema']['rv64']['schema']['ueie'][
        'default_setter'] = nusetter
    schema_yaml['mie']['schema']['rv32']['schema']['utie'][
        'default_setter'] = nusetter
    schema_yaml['mie']['schema']['rv64']['schema']['utie'][
        'default_setter'] = nusetter
    schema_yaml['mie']['schema']['rv32']['schema']['usie'][
        'default_setter'] = nusetter
    schema_yaml['mie']['schema']['rv64']['schema']['usie'][
        'default_setter'] = nusetter

    schema_yaml['mstatus']['schema']['rv32']['schema']['sie'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['sie'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv32']['schema']['fs'][
        'default_setter'] = fssetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['fs'][
        'default_setter'] = fssetter
    schema_yaml['mstatus']['schema']['rv32']['schema']['sd'][
        'default_setter'] = fssetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['sd'][
        'default_setter'] = fssetter
    schema_yaml['mstatus']['schema']['rv32']['schema']['spie'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['spie'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv32']['schema']['spp'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['spp'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv32']['schema']['tvm'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['tvm'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['sxl'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv32']['schema']['tsr'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['tsr'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv32']['schema']['mxr'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['mxr'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv32']['schema']['sum'][
        'default_setter'] = ssetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['sum'][
        'default_setter'] = ssetter
    schema_yaml['mip']['schema']['rv32']['schema']['seip'][
        'default_setter'] = ssetter
    schema_yaml['mip']['schema']['rv64']['schema']['seip'][
        'default_setter'] = ssetter
    schema_yaml['mip']['schema']['rv32']['schema']['stip'][
        'default_setter'] = ssetter
    schema_yaml['mip']['schema']['rv64']['schema']['stip'][
        'default_setter'] = ssetter
    schema_yaml['mip']['schema']['rv32']['schema']['ssip'][
        'default_setter'] = ssetter
    schema_yaml['mip']['schema']['rv64']['schema']['ssip'][
        'default_setter'] = ssetter
    schema_yaml['mip']['schema']['rv32']['schema']['vssip'][
        'default_setter'] = ssetter
    schema_yaml['mip']['schema']['rv64']['schema']['vssip'][
        'default_setter'] = ssetter
    schema_yaml['mip']['schema']['rv32']['schema']['vstip'][
        'default_setter'] = ssetter
    schema_yaml['mip']['schema']['rv64']['schema']['vstip'][
        'default_setter'] = ssetter
    schema_yaml['mip']['schema']['rv32']['schema']['vseip'][
        'default_setter'] = ssetter
    schema_yaml['mip']['schema']['rv64']['schema']['vseip'][
        'default_setter'] = ssetter
    schema_yaml['mip']['schema']['rv32']['schema']['sgeip'][
        'default_setter'] = hsetter
    schema_yaml['mip']['schema']['rv64']['schema']['sgeip'][
        'default_setter'] = hsetter
    schema_yaml['mie']['schema']['rv32']['schema']['seie'][
        'default_setter'] = ssetter
    schema_yaml['mie']['schema']['rv64']['schema']['seie'][
        'default_setter'] = ssetter
    schema_yaml['mie']['schema']['rv32']['schema']['stie'][
        'default_setter'] = ssetter
    schema_yaml['mie']['schema']['rv64']['schema']['stie'][
        'default_setter'] = ssetter
    schema_yaml['mie']['schema']['rv32']['schema']['ssie'][
        'default_setter'] = ssetter
    schema_yaml['mie']['schema']['rv64']['schema']['ssie'][
        'default_setter'] = ssetter
    schema_yaml['mie']['schema']['rv32']['schema']['vssie'][
        'default_setter'] = ssetter
    schema_yaml['mie']['schema']['rv64']['schema']['vssie'][
        'default_setter'] = ssetter
    schema_yaml['mie']['schema']['rv32']['schema']['vstie'][
        'default_setter'] = ssetter
    schema_yaml['mie']['schema']['rv64']['schema']['vstie'][
        'default_setter'] = ssetter
    schema_yaml['mie']['schema']['rv32']['schema']['vseie'][
        'default_setter'] = ssetter
    schema_yaml['mie']['schema']['rv64']['schema']['vseie'][
        'default_setter'] = ssetter
    schema_yaml['mie']['schema']['rv32']['schema']['sgeie'][
        'default_setter'] = hsetter
    schema_yaml['mie']['schema']['rv64']['schema']['sgeie'][
        'default_setter'] = hsetter
                
    schema_yaml['mstatus']['schema']['rv32']['schema']['tw'][
        'default_setter'] = twsetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['tw'][
        'default_setter'] = twsetter
    schema_yaml['mstatush']['schema']['rv32']['schema']['gva'][
        'default_setter'] = hsetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['gva'][
        'default_setter'] = hsetter
    schema_yaml['mstatush']['schema']['rv32']['schema']['mpv'][
        'default_setter'] = hsetter
    schema_yaml['mstatus']['schema']['rv64']['schema']['mpv'][
        'default_setter'] = hsetter
    schema_yaml['medeleg']['default_setter'] = delegsetter
    schema_yaml['mideleg']['default_setter'] = delegsetter
    schema_yaml['sedeleg']['default_setter'] = nregsetter
    schema_yaml['sideleg']['default_setter'] = nregsetter
    
    schema_yaml['hstatus']['schema']['rv32']['schema']['gva'][
        'default_setter'] = hsetter
    schema_yaml['hstatus']['schema']['rv64']['schema']['gva'][
        'default_setter'] = hsetter
    schema_yaml['hstatus']['schema']['rv32']['schema']['spv'][
        'default_setter'] = hsetter
    schema_yaml['hstatus']['schema']['rv64']['schema']['spv'][
        'default_setter'] = hsetter
    schema_yaml['hstatus']['schema']['rv32']['schema']['spvp'][
        'default_setter'] = hsetter
    schema_yaml['hstatus']['schema']['rv64']['schema']['spvp'][
        'default_setter'] = hsetter
    schema_yaml['hstatus']['schema']['rv32']['schema']['hu'][
        'default_setter'] = hsetter
    schema_yaml['hstatus']['schema']['rv64']['schema']['hu'][
        'default_setter'] = hsetter
    
    schema_yaml['hstatus']['default_setter'] = hregsetter
    schema_yaml['hedeleg']['default_setter'] = hregsetter
    schema_yaml['hideleg']['default_setter'] = hregsetter
    schema_yaml['hie']['default_setter'] = hregsetter
    schema_yaml['hip']['default_setter'] = hregsetter
    schema_yaml['hvip']['default_setter'] = hregsetter
    schema_yaml['hgeip']['default_setter'] = hregsetter
    schema_yaml['hgeie']['default_setter'] = hregsetter
    schema_yaml['htval']['default_setter'] = hregsetter
    schema_yaml['htinst']['default_setter'] = hregsetter
    schema_yaml['hgatp']['default_setter'] = hregsetter
    schema_yaml['htimedelta']['default_setter'] = hregsetter
    schema_yaml['htimedeltah']['default_setter'] = hreghsetter
    schema_yaml['hcounteren']['default_setter'] = uregsetter
    
    schema_yaml['hie']['schema']['rv32']['schema']['sgeie'][
        'default_setter'] = hsetter
    schema_yaml['hie']['schema']['rv64']['schema']['sgeie'][
        'default_setter'] = hsetter
    schema_yaml['hip']['schema']['rv32']['schema']['sgeip'][
        'default_setter'] = hsetter
    schema_yaml['hip']['schema']['rv64']['schema']['sgeip'][
        'default_setter'] = hsetter
    
    schema_yaml['vsstatus']['default_setter'] = sregsetter
    schema_yaml['vsstatus']['schema']['rv32']['schema']['uie'][
        'default_setter'] = nusetter
    schema_yaml['vsstatus']['schema']['rv64']['schema']['uie'][
        'default_setter'] = nusetter
    schema_yaml['vsstatus']['schema']['rv32']['schema']['upie'][
        'default_setter'] = nusetter
    schema_yaml['vsstatus']['schema']['rv64']['schema']['upie'][
        'default_setter'] = nusetter

    schema_yaml['vsstatus']['schema']['rv64']['schema']['uxl'][
        'default_setter'] = usetter
    schema_yaml['vsstatus']['schema']['rv32']['schema']['sie'][
        'default_setter'] = ssetter
    schema_yaml['vsstatus']['schema']['rv64']['schema']['sie'][
        'default_setter'] = ssetter
    schema_yaml['vsstatus']['schema']['rv32']['schema']['spie'][
        'default_setter'] = ssetter
    schema_yaml['vsstatus']['schema']['rv64']['schema']['spie'][
        'default_setter'] = ssetter
    schema_yaml['vsstatus']['schema']['rv32']['schema']['spp'][
        'default_setter'] = ssetter
    schema_yaml['vsstatus']['schema']['rv64']['schema']['spp'][
        'default_setter'] = ssetter
    schema_yaml['vsstatus']['schema']['rv32']['schema']['mxr'][
        'default_setter'] = ssetter
    schema_yaml['vsstatus']['schema']['rv64']['schema']['mxr'][
        'default_setter'] = ssetter
    schema_yaml['vsstatus']['schema']['rv32']['schema']['sum'][
        'default_setter'] = ssetter
    schema_yaml['vsstatus']['schema']['rv64']['schema']['sum'][
        'default_setter'] = ssetter
    schema_yaml['vsie']['default_setter'] = sregsetter
    schema_yaml['vsie']['schema']['rv32']['schema']['seie'][
        'default_setter'] = ssetter
    schema_yaml['vsie']['schema']['rv64']['schema']['seie'][
        'default_setter'] = ssetter
    schema_yaml['vsie']['schema']['rv32']['schema']['stie'][
        'default_setter'] = ssetter
    schema_yaml['vsie']['schema']['rv64']['schema']['stie'][
        'default_setter'] = ssetter
    schema_yaml['vsie']['schema']['rv32']['schema']['ssie'][
        'default_setter'] = ssetter
    schema_yaml['vsie']['schema']['rv64']['schema']['ssie'][
        'default_setter'] = ssetter
    schema_yaml['vsip']['default_setter'] = sregsetter
    schema_yaml['vsip']['schema']['rv32']['schema']['seip'][
        'default_setter'] = ssetter
    schema_yaml['vsip']['schema']['rv64']['schema']['seip'][
        'default_setter'] = ssetter
    schema_yaml['vsip']['schema']['rv32']['schema']['stip'][
        'default_setter'] = ssetter
    schema_yaml['vsip']['schema']['rv64']['schema']['stip'][
        'default_setter'] = ssetter
    schema_yaml['vsip']['schema']['rv32']['schema']['ssip'][
        'default_setter'] = ssetter
    schema_yaml['vsip']['schema']['rv64']['schema']['ssip'][
        'default_setter'] = ssetter
    schema_yaml['vstvec']['default_setter'] = sregsetter
    schema_yaml['vsepc']['default_setter'] = sregsetter
    schema_yaml['vstval']['default_setter'] = sregsetter
    schema_yaml['vscause']['default_setter'] = sregsetter
    schema_yaml['vsatp']['default_setter'] = sregsetter
    schema_yaml['vsscratch']['default_setter'] = sregsetter
    return schema_yaml


def trim(foo):
    '''
        Function to trim the dictionary. Any node with implemented field set to false is trimmed of all the other nodes.

        :param foo: The dictionary to be trimmed.

        :type foo: dict

        :return: The trimmed dictionary.
    '''
    keys = foo.keys()
    if 'rv32' in keys:
        if not foo['rv32']['accessible']:
            foo['rv32'] = {'accessible': False}
    if 'rv64' in keys:
        if not foo['rv64']['accessible']:
            foo['rv64'] = {'accessible': False}
    if 'implemented' in keys:
        if not foo['implemented']:
            temp = foo
            for k in list(
                    set(foo.keys()) -
                    set(['description', 'msb', 'lsb', 'implemented', 'shadow', 'shadow_type'])
            ):
                try:
                    temp.pop(k)
                except KeyError:
                    continue
            return temp
    for key in keys:
        if isinstance(foo[key], dict):
            t = trim(foo[key])
            foo[key] = t
    return foo


def groupc(test_list):
    ''' Generator function to squash consecutive numbers for wpri bits.'''
    for x, y in itertools.groupby(enumerate(test_list), lambda a: a[1] - a[0]):
        y = list(y)
        a = y[0][1]
        b = y[-1][1]
        if a == b:
            yield [a]
        else:
            yield [a, b]


def get_fields(node, bitwidth):
    fields = list(
        set(node.keys()) -
        set(['fields', 'msb', 'lsb', 'accessible', 'shadow', 'shadow_type','type']))

    if not fields:
        return fields
    nf = {}
    for x in fields:
        nf[x] = node[x]['lsb']
    nf = sorted(nf.items(), key=lambda x: x[1], reverse=False)
    nfields=[]
    for k, v in nf:
        nfields.append(k)

    fields = nfields
    bits = set(range(bitwidth))
    for entry in fields:
        bits -= set(range(node[entry]['lsb'], node[entry]['msb'] + 1))
    bits = list(groupc(sorted(list(bits))))
    if not bits:
        return fields
    else:
        fields.append(bits)
        return fields
        
def check_fields(spec):
    errors = {} 
    for csr, node, in spec.items() :
         fault_node = node
         error=[]
         if node['rv32']['accessible']:
                node['rv32']['fields'] = get_fields(node['rv32'], 32)
         if node['rv64']['accessible']:
                node['rv64']['fields'] = get_fields(node['rv64'], 64)
         fields = list(set(['rv32', 'rv64', 'description', 'address', 'priv_mode', 'reset-val']) - set(node.keys()) )
         if fields:
            error.append("The fields " + "".join(fields) + " are missing")
         if node['rv32']['accessible']:
            if any(type(e)==list for e in node['rv32']['fields']): 
             sub_fields = node['rv32']['fields'][:-1]
            else:
             sub_fields = node['rv32']['fields']
            if not sub_fields :
             subfields = list(set(['msb', 'lsb', 'accessible', 'shadow', 'shadow_type', 'fields', 'type']) - set(node['rv32'].keys()) )    
             if subfields:
                error.append("The subfield " + "".join(subfields) + " are not present")         
            else:
              for x in sub_fields :
                subfields = list(set(['msb', 'lsb', 'implemented', 'description', 'shadow', 'shadow_type', 'type']) - set(node['rv32'][x].keys()) )
                if subfields :                   
                   error.append("The subfields " + "".join(subfields) + " are not present in " + str(x))
         if node['rv64']['accessible']:            
            if any(type(e)==list for e in node['rv64']['fields']): 
             sub_fields = node['rv64']['fields'][:-1]
            else:
             sub_fields = node['rv64']['fields']
            if not sub_fields :
             subfields = list(set(['msb', 'lsb', 'accessible', 'fields', 'shadow', 'shadow_type', 'type']) - set(node['rv64'].keys()))
             if subfields:
                error.append("The subfield " + "".join(subfields) + " are not present")
            else:
              for x in sub_fields :
                subfields = list(set(['msb', 'lsb', 'implemented', 'description', 'shadow', 'shadow_type', 'type']) - set(node['rv64'][x].keys()) )
                if subfields :                   
                   error.append("The subfields " + "".join(subfields) + " are not present in " + str(x))
         if bin(node['address'])[2:][::-1][6:8] != '11' and bin(node['address'])[2:][::-1][8:12] != '0001':
             error.append('Address is not in custom csr ranges')
         if (bin(node['address'])[2:][::-1][8:10] == '00' and node['priv_mode'] != 'U' ) or (bin(node['address'])[2:][::-1][8:10] == '01' and node['priv_mode'] != 'S' ) or (bin(node['address'])[2:][::-1][8:10] == '11' and node['priv_mode'] != 'M') :
            error.append('Privilege does not match with the address')
         if error:
            errors[csr] = error
    return errors 
def check_shadows(spec, logging = False):
    ''' Check if the shadowed fields are implemented and of the same size as the
    source'''
    errors = {}
    _rvxlen = ['rv32', 'rv64']
    for csr, content in spec.items():
        if logging:
            logger.debug('Checking Shadows for ' + csr)
        error = []
        if isinstance(content, dict) and 'description' in content:
            for rvxlen in _rvxlen:
                if content[rvxlen]['accessible'] and not content[rvxlen]['fields']:
                    if content[rvxlen]['shadow'] is None: 
                        continue
                    else:
                        shadow = content[rvxlen]['shadow'].split('.')
                        if len(shadow) > 2:
                            error.append('Shadow field of should only up to 1 dot')
                            continue
                        elif len(shadow) == 2:
                            scsr = shadow[0]
                            subscsr = shadow[1]
                            if not spec[scsr][rvxlen][subscsr]['implemented']:
                                error.append(' Shadow field ' + scsr + '.' +\
                                            subscsr + ' not implemented')
                                continue
                            scsr_size = spec[scsr][rvxlen][subscsr]['msb'] -\
                                    spec[scsr][rvxlen][subscsr]['lsb']
                            csr_size = spec[csr][rvxlen]['msb'] - spec[csr][rvxlen]['lsb']
                            if scsr_size != csr_size :
                                error.append('Shadow field '+ \
                                        scsr + '.' + subscsr + \
                                        ' does not match in size')
                        else:
                            scsr = shadow[0]
                            if not spec[scsr][rvxlen]['accessible']:
                                error.append('Shadow field ' + scsr + ' not implemented')
                                continue
                            scsr_size = spec[scsr][rvxlen]['msb'] - spec[scsr][rvxlen]['lsb']
                            csr_size = spec[csr][rvxlen]['msb'] - spec[csr][rvxlen]['lsb']
                            if scsr_size != csr_size :
                                error.append('Shadow field '+ scsr +\
                                        'does not match in size') 
                elif content[rvxlen]['accessible']:
                    for subfield in content[rvxlen]['fields']:
                        if isinstance(subfield ,list):
                            continue
                        if content[rvxlen][subfield]['shadow'] is None:
                           continue
                        elif content[rvxlen][subfield]['implemented']:
                            shadow = content[rvxlen][subfield]['shadow'].split('.')
                            if len(shadow) > 2:
                                error.append('Shadow field of should only up to 1 dot')
                                continue
                            elif len(shadow) == 1:
                                scsr = shadow[0]
                                if not spec[scsr][rvxlen]['accessible']:
                                    error.append('Subfield ' + subfield + \
                                        ' shadowing ' + scsr + ' not implemented')
                                    continue
                                scsr_size = spec[scsr][rvxlen]['msb'] - spec[scsr][rvxlen]['lsb']
                                csr_size = spec[csr][rvxlen][subfield]['msb'] -\
                                        spec[csr][rvxlen][subfield]['lsb']
                                if scsr_size != csr_size :
                                    error.append('Subfield ' + subfield +'shadowing'+ \
                                            scsr + ' does not match in size')
                            else:
                                scsr = shadow[0]
                                subscsr = shadow[1]
                                if not spec[scsr][rvxlen][subscsr]['implemented']:
                                    error.append('Subfield ' + subfield + \
                                            ' shadowing ' + scsr + '.' +\
                                            subscsr + ' not implemented')
                                    continue
                                scsr_size = spec[scsr][rvxlen][subscsr]['msb'] -\
                                        spec[scsr][rvxlen][subscsr]['lsb']
                                csr_size = spec[csr][rvxlen][subfield]['msb'] -\
                                        spec[csr][rvxlen][subfield]['lsb']
                                if scsr_size != csr_size :
                                    error.append('Subfield ' + subfield +'shadowing'+ \
                                            scsr + '.' + subscsr + \
                                            ' does not match in size') 

        if error:
            errors[csr] = error
    return errors

def check_mhpm(spec, logging = False):
    ''' Check if the mhpmcounters and corresponding mhpmevents are implemented and of the same size as the
    source'''
    errors = {}
    for csrname, content, in spec.items():
        error = []
        if 'mhpmcounter' in csrname:
            index = int(re.findall('\d+',csrname.lower())[0])
            if content['rv64']['accessible'] :
                if not spec['mhpmevent'+str(index)]['rv64']['accessible']:
                    error.append(csrname + " counter doesn't have the corresponding mhpmevent register accessible")
            if content['rv32']['accessible'] :
                if not spec['mhpmevent'+str(index)]['rv32']['accessible']:
                    error.append(csrname + " counter doesn't have the corresponding mhpmevent register accessible")
        if 'mhpmevent' in csrname:
            index = int(re.findall('\d+',csrname.lower())[0])
            if content['rv64']['accessible'] :
                if not spec['mhpmcounter'+str(index)]['rv64']['accessible']:
                    error.append(csrname + " event reg doesn't have the corresponding mhpmcounter register accessible")
            if content['rv32']['accessible'] :
                if not spec['mhpmcounter'+str(index)]['rv32']['accessible']:
                    error.append(csrname + " event reg doesn't have the corresponding mhpmcounter register accessible")
            if content['rv32']['accessible'] :
                if not spec['mhpmcounter'+str(index)+'h']['rv32']['accessible']:
                    error.append(csrname + " event reg doesn't have the corresponding mhpmcounter 'h' counterpart register accessible")
        if error:
            errors[csrname] = error
    return errors
    
     
def check_pmp(spec, logging = False):
    ''' Check if the mhpmcounters and corresponding mhpmevents are implemented and of the same size as the
    source'''
    errors = {}
    for csrname, content, in spec.items():
        error = []
        Grain=int(spec['pmp_granularity'])
        if 'pmpaddr' in csrname:
            index = int(re.findall('\d+',csrname.lower())[0])
            if content['rv64']['accessible'] :                
                reset_val_addr = (bin(content['reset-val'])[2:].zfill(64))[::-1] 
                reset_val_cfg  = (bin(spec['pmpcfg'+str(int(int(index/8)*2))]['reset-val'])[2:].zfill(64))[::-1]
                if not spec['pmpcfg'+str(int(int(index/8)*2))]['rv64']['accessible']:
                    error.append(csrname + " addr doesn't have the corresponding pmp config register accessible")
                if not spec['pmpcfg'+str(int(int(index/8)*2))]['rv64']['pmp'+str(index)+'cfg']['implemented'] :
                    error.append(csrname + " addr doesn't have the corresponding pmpcfg" +str(int(index/4)) + "_pmp" + str(index) +"cfg register implemented")
                if reset_val_cfg[8*(index-(int(index/8)*8)) + 4] == '1' and Grain >=2 :     #NAPOT, Bit A of pmpXcfg is set
                  if '0' in reset_val_addr[0:(Grain-1)] or (reset_val_addr[Grain-1] != '0') :
                    error.append(csrname + 'reset value does not adhere with the pmp granularity')
                elif Grain >= 1: #TOR
                  if int(content['reset-val']) % (2**Grain) != 0 :
                    error.append(csrname + 'reset value does not adhere with the pmp granularity')
            if content['rv32']['accessible'] :
                reset_val_addr = (bin(content['reset-val'])[2:].zfill(32))[::-1] 
                reset_val_cfg  = (bin(spec['pmpcfg'+str(int(index/4))]['reset-val'])[2:].zfill(32))[::-1]
                if not spec['pmpcfg'+str(int(index/4))]['rv32']['accessible']:
                    error.append(csrname + " addr doesn't have the corresponding pmp config register accessible")
                if not spec['pmpcfg'+str(int(index/4))]['rv32']['pmp'+str(index)+'cfg']['implemented'] :
                    error.append(csrname + " addr doesn't have the corresponding pmpcfg" +str(int(index/4)) + "_pmp" + str(index) +"cfg register implemented")
                if reset_val_cfg[8*(index-(int(index/4)*4)) + 4] == '1' and Grain >=2 :     #NAPOT, Bit A of pmpXcfg is set
                 if '0' in reset_val_addr[0:(Grain-1)] or (reset_val_addr[Grain-1] != '0') :
                    error.append(csrname + 'reset value does not adhere with the pmp granularity')
                elif Grain >= 1: #TOR
                  if int(content['reset-val']) % (2**Grain) != 0 :
                    error.append(csrname + 'reset value does not adhere with the pmp granularity')
        if 'pmpcfg' in csrname:
            if content['rv64']['accessible'] :
                for subfield in content['rv64']['fields']:
                 index = int(re.findall('\d+',subfield)[0])
                 if content['rv64'][subfield]['implemented'] and not spec['pmpaddr'+str(index)]['rv64']['accessible']:
                    error.append(csrname + "_" + subfield + " doesn't have the corresponding pmpaddr accessible")
            if content['rv32']['accessible'] :
                for subfield in content['rv32']['fields']:
                 index = int(re.findall('\d+',subfield)[0])
                 if content['rv32'][subfield]['implemented'] and not spec['pmpaddr'+str(index)]['rv32']['accessible']:
                    error.append(csrname + "_" + subfield + " doesn't have the corresponding pmpaddr accessible")
        if error:
            errors[csrname] = error
    return errors
   





def check_reset_fill_fields(spec, logging= False):
    '''The check_reset_fill_fields function fills the field node with the names of the sub-fields of the register and then checks whether the reset-value of the register is a legal value. To do so, it iterates over all the subfields and extracts the corresponding field value from the reset-value. Then it checks the legality of the value according to the given field description. If the fields is implemented i.e accessible in both 64 bit and 32 bit modes, the 64 bit mode is given preference. '''
    errors = {}
    for node in spec:

        if isinstance(spec[node], dict):
            if logging:
                logger.debug('Checking Reset fields for: ' +  node)
            if spec[node]['rv32']['accessible']:
                spec[node]['rv32']['fields'] = get_fields(
                    spec[node]['rv32'], 32)
            if spec[node]['rv64']['accessible']:
                spec[node]['rv64']['fields'] = get_fields(
                    spec[node]['rv64'], 64)
            if 'reset-val' in spec[node].keys():
                reset_val = spec[node]['reset-val']
                if spec[node]['rv64']['accessible']:
                    field_desc = spec[node]['rv64']
                    bit_len = 64
                elif spec[node]['rv32']['accessible']:
                    field_desc = spec[node]['rv32']
                    bit_len = 32
                else:
                    continue
                error = []
                if not field_desc['fields']:
                    if field_desc['shadow'] is None:
                        desc = field_desc['type']
                        keys = desc.keys()
                        if 'wlrl' in keys:
                            res = False
                            for entry in desc['wlrl']:
                                if ":" in entry:
                                    low, high = entry.split(":")
                                    if "x" in low:
                                        low = int(low, base=16)
                                        high = int(high, base=16)
                                    else:
                                        low = int(low)
                                        high = int(high)
                                    if reset_val >= low and reset_val <= high:
                                        res = True
                                        break
                                else:
                                    if "x" in entry:
                                        val = int(entry, base=16)
                                    else:
                                        val = int(entry)
                                    if val == reset_val:
                                        res = True
                                        break
                            if not res:
                                error.append(
                                    "Reset value doesnt match the 'wlrl' description for the register."
                                )
                        elif 'ro_constant' in keys:
                            if reset_val != desc['ro_constant']:
                                error.append(
                                    "Reset value doesnt match the 'ro_constant' description for the register."
                                )
                        elif 'ro_variable' in keys:
                            pass
                        elif "warl" in keys:
                            warl = (warl_interpreter(desc['warl']))
                            deps = warl.dependencies
                            dep_vals = []
                            for dep in deps:
                                reg = dep.split("::")
                                if len(reg) == 1:
                                    dep_vals.append(spec[reg[0]]['reset-val'])
                                else:
                                    bin_str = bin(spec[reg[0]]
                                                  ['reset-val'])[2:].zfill(bit_len)
                                    msb = spec[reg[0]]['rv{}'.format(bit_len)][reg[1]]['msb']
                                    lsb = spec[reg[0]]['rv{}'.format(bit_len)][reg[1]]['lsb']
                                    dep_vals.append(
                                        int(bin_str[::-1][lsb:msb+1][::-1], base=2))
                            if (warl.islegal(int(reset_val), dep_vals) != True):
                                error.append(
                                    "Reset value doesnt match the 'warl' description for the register."
                                )
                else:
                    bin_str = bin(reset_val)[2:].zfill(bit_len)
                    for field in field_desc['fields']:
                        if isinstance(field, list):
                            for entry in field:
                                if len(entry) == 2:
                                    test_val = int(
                                        bin_str[bit_len - 1 - entry[1]:bit_len -
                                                entry[0]],
                                        base=2)
                                    if test_val != 0:
                                        error.append("WPRI bits " +
                                                     str(entry[0]) + " to " +
                                                     str(entry[1]) +
                                                     " should be zero.")
                                elif len(entry) == 1:
                                    test_val = int(bin_str[bit_len - 1 -
                                                           entry[0]],
                                                   base=2)
                                    if test_val != 0:
                                        error.append("WPRI bit " +
                                                     str(entry[0]) +
                                                     " should be zero.")
                        else:
                            test_val = int(
                                bin_str[bit_len - 1 -
                                        field_desc[field]['msb']:bit_len -
                                        field_desc[field]['lsb']],
                                base=2)
                            if field_desc[field]['implemented']:
                                if field_desc[field]['shadow'] is None:
                                    logger.debug('--> Subfield: ' + field)
                                    desc = field_desc[field]['type']
                                    keys = desc.keys()
                                    if 'wlrl' in keys:
                                        res = False
                                        for entry in desc['wlrl']:
                                            if ":" in entry:
                                                low, high = entry.split(":")
                                                if "x" in low:
                                                    low = int(low, base=16)
                                                    high = int(high, base=16)
                                                else:
                                                    low = int(low)
                                                    high = int(high)
                                                if test_val >= low and test_val <= high:
                                                    res = True
                                                    break
                                            else:
                                                if "x" in entry:
                                                    val = int(entry, base=16)
                                                else:
                                                    val = int(entry)
                                                if val == test_val:
                                                    res = True
                                                    break
                                        if not res:
                                            error.append(
                                                "Reset value for " + field +
                                                " doesnt match the 'wlrl' description."
                                            )
                                    elif 'ro_constant' in keys:
                                        if test_val != desc['ro_constant']:
                                            error.append(
                                                "Reset value for " + field +
                                                " doesnt match the 'ro_constant' description."
                                            )
                                    elif 'ro_variable' in keys:
                                        pass
                                    elif "warl" in keys:
                                        warl = (warl_interpreter(desc['warl']))
                                        deps = warl.dependencies
                                        dep_vals = []
                                        for dep in deps:
                                            reg = dep.split("::")
                                            if len(reg) == 1:
                                                dep_vals.append(
                                                    spec[reg[0]]['reset-val'])
                                            else:
                                                bin_str = bin(
                                                    spec[reg[0]]['reset-val']
                                                )[2:].zfill(bit_len)
                                                dep_vals.append(
                                                    int(bin_str[
                                                        bit_len - 1 -
                                                        spec[reg[0]]['rv{}'.format(
                                                            bit_len
                                                        )][reg[1]]['msb']:bit_len -
                                                        spec[reg[0]]['rv{}'.format(
                                                            bit_len
                                                        )][reg[1]]['lsb']],
                                                        base=2))
                                        if (warl.islegal(
                                                int(test_val), dep_vals) !=
                                                True):
                                            error.append(
                                                "Reset value for " + field +
                                                " doesnt match the 'warl' description."
                                            )
                            elif test_val != 0:
                                error.append("Reset value for unimplemented " +
                                             field + " cannot be non zero.")
                if error:
                    errors[node] = error
    return spec, errors

def check_debug_specs(debug_spec, isa_spec,
                work_dir,
                logging=False,
                no_anchors=False):
    '''
        Function to perform ensure that the isa and debug specifications confirm
        to their schemas. The :py:mod:`Cerberus` module is used to validate that the
        specifications confirm to their respective schemas.


        :param debug_spec: The path to the DUT debug specification yaml file.

        :param isa_spec: The path to the DUT isa specification yaml file.

        :param logging: A boolean to indicate whether log is to be printed.

        :type logging: bool

        :type isa_spec: str

        :raise ValidationError: It is raised when the specifications violate the
            schema rules. It also contains the specific errors in each of the fields.

        :return: A tuple with the first entry being the absolute path to normalized isa file
            and the second being the absolute path to the platform spec file.
    '''

    if logging:
        logger.info('Input-Debug file')
    
    foo1 = isa_spec
    foo = debug_spec
    schema = constants.debug_schema
    """
      Read the input-isa foo (yaml file) and validate with schema-isa for feature values
      and constraints
    """
    # Load input YAML file
    if logging:
        logger.info('Loading input file: ' + str(foo))
    master_inp_debug_yaml = utils.load_yaml(foo, no_anchors)

    # Load input YAML file
    if logging:
        logger.info('Loading input isa file: ' + str(foo1))
    master_inp_yaml = utils.load_yaml(foo1, no_anchors)
    isa_string = master_inp_yaml['hart0']['ISA']
    
    # instantiate validator
    if logging:
        logger.info('Load Schema ' + str(schema))
    master_schema_yaml = utils.load_yaml(schema, no_anchors)

    outyaml = copy.deepcopy(master_inp_debug_yaml)
    for x in master_inp_debug_yaml['hart_ids']:
        if logging:
            logger.info('Processing Hart: hart'+str(x))
        inp_debug_yaml = master_inp_debug_yaml['hart'+str(x)]
        schema_yaml = add_debug_setters(master_schema_yaml['hart_schema']['schema'])
        #Extract xlen
        xlen = inp_debug_yaml['supported_xlen']

        validator = schemaValidator(schema_yaml, xlen=xlen, isa_string=isa_string)
        validator.allow_unknown = False
        validator.purge_readonly = True
        normalized = validator.normalized(inp_debug_yaml, schema_yaml)

        # Perform Validation
        if logging:
            logger.info('Initiating Validation')
        valid = validator.validate(normalized)

        # Print out errors
        if valid:
            if logging:
                logger.info('No errors for Hart: '+str(x) + ' :)')
        else:
            error_list = validator.errors
            raise ValidationError("Error in " + foo + ".", error_list)
        if logging:
            logger.info("Initiating post processing and reset value checks.")
        normalized, errors = check_reset_fill_fields(normalized, logging)
        if errors:
            raise ValidationError("Error in " + foo + ".", errors)
        outyaml['hart'+str(x)] = trim(normalized)
    file_name = os.path.split(foo)
    file_name_split = file_name[1].split('.')
    output_filename = os.path.join(
        work_dir, file_name_split[0] + '_checked.' + file_name_split[1])
    ifile = output_filename
    outfile = open(output_filename, 'w')
    if logging:
        logger.info('Dumping out Normalized Checked YAML: ' + output_filename)
    utils.dump_yaml(outyaml, outfile, no_anchors )
    return ifile
    
def check_isa_specs(isa_spec,
                work_dir,
                logging=False,
                no_anchors=False):
    '''
        Function to perform ensure that the isa and platform specifications confirm
        to their schemas. The :py:mod:`Cerberus` module is used to validate that the
        specifications confirm to their respective schemas.

        :param isa_spec: The path to the DUT isa specification yaml file.

        :param logging: A boolean to indicate whether log is to be printed.

        :type logging: bool

        :type isa_spec: str

        :raise ValidationError: It is raised when the specifications violate the
            schema rules. It also contains the specific errors in each of the fields.

        :return: A tuple with the first entry being the absolute path to normalized isa file
            and the second being the absolute path to the platform spec file.
    '''
    global inp_yaml

    if logging:
        logger.info('Input-ISA file')

    foo = isa_spec
    schema = constants.isa_schema
    """
      Read the input-isa foo (yaml file) and validate with schema-isa for feature values
      and constraints
    """
    # Load input YAML file
    if logging:
        logger.info('Loading input file: ' + str(foo))
    master_inp_yaml = utils.load_yaml(foo, no_anchors)

    # instantiate validator
    if logging:
        logger.info('Load Schema ' + str(schema))
    master_schema_yaml = utils.load_yaml(schema, no_anchors)

    outyaml = copy.deepcopy(master_inp_yaml)
    for x in master_inp_yaml['hart_ids']:
        if logging:
            logger.info('Processing Hart: hart'+str(x))
        inp_yaml = master_inp_yaml['hart'+str(x)]
        schema_yaml = add_def_setters(master_schema_yaml['hart_schema']['schema'])
        schema_yaml = add_reset_setters(master_schema_yaml['hart_schema']['schema']) 
        schema_yaml = add_fflags_type_setters(master_schema_yaml['hart_schema']['schema']) 
        #Extract xlen
        xlen = inp_yaml['supported_xlen']
        rvxlen='rv'+str(xlen[0])
        validator = schemaValidator(schema_yaml, xlen=xlen, isa_string=inp_yaml['ISA'])
        validator.allow_unknown = False
        validator.purge_readonly = True
        normalized = validator.normalized(inp_yaml, schema_yaml)

        # Perform Validation
        if logging:
            logger.info('Initiating Validation')
        valid = validator.validate(normalized)

        # Print out errors
        if valid:
            if logging:
                logger.info('No errors for Hart: '+str(x) + ' :)')
        else:
            error_list = validator.errors
            raise ValidationError("Error in " + foo + ".", error_list)
        if logging:
            logger.info("Initiating post processing and reset value checks.")
        normalized, errors = check_reset_fill_fields(normalized, logging)
        if errors:
            raise ValidationError("Error in " + foo + ".", errors)
        if normalized['mhartid']['reset-val'] != x:
            raise ValidationError('Error in ' + foo + ".", 
                    {'mhartid': ['wrong reset-val of for hart'+str(x)]})
        errors = check_shadows(normalized, logging)
        if errors:
            raise ValidationError("Error in " + foo + ".", errors)
        errors = check_mhpm(normalized, logging)
        if errors:
            raise ValidationError("Error in " + foo + ".", errors)
        errors = check_pmp(normalized, logging)
        if errors:
            raise ValidationError("Error in " + foo + ".", errors) 
              
        outyaml['hart'+str(x)] = trim(normalized)
    file_name = os.path.split(foo)
    file_name_split = file_name[1].split('.')
    output_filename = os.path.join(
        work_dir, file_name_split[0] + '_checked.' + file_name_split[1])
    ifile = output_filename
    outfile = open(output_filename, 'w')
    if logging:
        logger.info('Dumping out Normalized Checked YAML: ' + output_filename)
    utils.dump_yaml(outyaml, outfile, no_anchors )
    return ifile
    
def check_custom_specs(custom_spec,
                work_dir,
                logging=False,
                no_anchors=False):
    '''
        Function to perform ensure that the isa and platform specifications confirm
        to their schemas. The :py:mod:`Cerberus` module is used to validate that the
        specifications confirm to their respective schemas.

        :param isa_spec: The path to the DUT isa specification yaml file.

        :param logging: A boolean to indicate whether log is to be printed.

        :type logging: bool

        :type isa_spec: str

        :raise ValidationError: It is raised when the specifications violate the
            schema rules. It also contains the specific errors in each of the fields.

        :return: A tuple with the first entry being the absolute path to normalized isa file
            and the second being the absolute path to the platform spec file.
    '''
    if logging:
        logger.info('Custom CSR Spec')

    foo = custom_spec
    
    # Load input YAML file
    if logging:
        logger.info('Loading input file: ' + str(foo))
    master_custom_yaml = utils.load_yaml(foo, no_anchors)

    outyaml = copy.deepcopy(master_custom_yaml)
    for x in master_custom_yaml['hart_ids']:
        if logging:
            logger.info('Processing Hart: hart'+str(x))
        inp_yaml = master_custom_yaml['hart'+str(x)]
    errors = check_fields(inp_yaml)
    if errors:
            raise ValidationError("Error in " + foo + ".", errors)
    outyaml['hart'+str(x)] = trim(inp_yaml)
    file_name = os.path.split(foo)
    file_name_split = file_name[1].split('.')
    output_filename = os.path.join(
        work_dir, file_name_split[0] + '_checked.' + file_name_split[1])
    cfile = output_filename
    outfile = open(output_filename, 'w')
    if logging:
        logger.info('Dumping out Normalized Checked YAML: ' + output_filename)
    utils.dump_yaml(outyaml, outfile, no_anchors )
    return cfile
    
def check_platform_specs(platform_spec,
                work_dir,
                logging=False,
                no_anchors=False):
    foo = platform_spec
    schema = constants.platform_schema
    if logging:
        logger.info('Input-Platform file')
    """
      Read the input-platform foo (yaml file) and validate with schema-platform for feature values
      and constraints
    """
    # Load input YAML file
    if logging:
        logger.info('Loading input file: ' + str(foo))
    inp_yaml = utils.load_yaml(foo, no_anchors)
    if inp_yaml is None:
        inp_yaml = {'mtime': {'implemented': False}}

    # instantiate validator
    if logging:
        logger.info('Load Schema ' + str(schema))
    schema_yaml = utils.load_yaml(schema, no_anchors)
    validator = schemaValidator(schema_yaml, xlen=[])
    validator.allow_unknown = False
    validator.purge_readonly = True
    normalized = validator.normalized(inp_yaml, schema_yaml)

    # Perform Validation
    if logging:
        logger.info('Initiating Validation')
    valid = validator.validate(normalized)

    # Print out errors
    if valid:
        if logging:
            logger.info('No Syntax errors in Input Platform Yaml. :)')
    else:
        error_list = validator.errors
        raise ValidationError("Error in " + foo + ".", error_list)

    file_name = os.path.split(foo)
    file_name_split = file_name[1].split('.')
    output_filename = os.path.join(
        work_dir, file_name_split[0] + '_checked.' + file_name_split[1])
    pfile = output_filename
    outfile = open(output_filename, 'w')
    if logging:
        logger.info('Dumping out Normalized Checked YAML: ' + output_filename)
    utils.dump_yaml(trim(normalized), outfile, no_anchors)

    return pfile



import riscv_config.constants as constants
import re

def get_extension_list(isa):
    extension_list = []
    err = False
    err_list = []
    if not constants.isa_regex.match(isa):
        err = True
        err_list.append('Input ISA string does not match accepted canonical ordering')
        return (extension_list, err, err_list)

    
    str_match = re.findall('(?P<stdisa>[^\d]*?)(?!_)*(?P<zext>Z.*?)*(?P<sext>S[a-z]*)*(_|$)',isa)
    extension_list= []
    standard_isa = ''
    zext_list = []
    for match in str_match:
        stdisa, zext, sext, ignore = match
        if stdisa != '':
            for e in stdisa:
                extension_list.append(e)
            standard_isa = stdisa
        if zext != '':
            extension_list.append(zext)
            zext_list.append(zext)
        if sext != '':
            extension_list.append(sext)
    # check ordering of ISA
    canonical_ordering = 'IEMAFDQLCBJKTPVNSHU'
    order_index = {c: i for i, c in enumerate(canonical_ordering)}
    for i in range(len(standard_isa)-1):
        a1 = standard_isa[i]
        a2 = standard_isa[i+1]
    
        if order_index[a1] > order_index[a2]:
            err = True
            err_list.append( "Alphabet '" + a1 + "' should occur after '" + a2)

    # check canonical ordering within Zextensions
    for i in range(len(zext_list)-1):
        a1 = zext_list[i][1].upper()
        a2 = zext_list[i+1][1].upper()
        a3 = zext_list[i][2]
        a4 = zext_list[i+1][2]
        if order_index[a1] > order_index[a2]:
            err = True
            err_list.append( f"Z extension {zext_list[i]} must occur after {zext_list[i+1]}")
        elif a1 == a2 and a3 > a4:
            err = True
            err_list.append( f"Within the Z{a1.lower()} category extension {zext_list[i]} must occur after {zext_list[i+1]}")
        
    if 'I' not in extension_list and 'E' not in extension_list:
        err_list.append( 'Either of I or E base extensions need to be present in the ISA string')
        err = True
    if 'F' in extension_list and not "Zicsr" in extension_list:
        err_list.append( "F cannot exist without Zicsr.")
        err = True
    if 'D' in extension_list and not 'F' in extension_list:
        err_list.append( "D cannot exist without F.")
        err = True
    if 'Q' in extension_list and not 'D' in extension_list:
        err_list.append( "Q cannot exist without D and F.")
        err = True
    if 'Zam' in extension_list and not 'A' in extension_list:
        err_list.append( "Zam cannot exist without A.")
        err = True
    if 'N' in extension_list and not 'U' in extension_list:
        err_list.append( "N cannot exist without U.")
        err = True
    if 'S' in extension_list and not 'U' in extension_list:
        err_list.append( "S cannot exist without U.")
        err = True
    if 'Zkn' in extension_list and ( set(['Zbkb', 'Zbkc', 'Zbkx', 'Zkne', 'Zknd', 'Zknh']) & set(extension_list)):
        err_list.append( "Zkn is a superset of Zbkb, Zbkc, Zbkx, Zkne, Zknd, Zknh. In presence of Zkn the subsets must be ignored in the ISA string.")
        err = True
    if 'Zks' in extension_list and ( set(['Zbkb', 'Zbkc', 'Zbkx', 'Zksed', 'Zksh']) & set(extension_list) ):
        err_list.append( "Zks is a superset of Zbkb, Zbkc, Zbkx, Zksed, Zksh. In presence of Zks the subsets must be ignored in the ISA string.")
        err = True
    if 'Zk' in extension_list and ( set(['Zbkb', 'Zbkc', 'Zbkx', 'Zkne', 'Zknd',
        'Zknh', 'Zkn','Zkr','Zkt']) & set(extension_list)):
        err_list.append( "Zkn is a superset of Zbkb, Zbkc, Zbkx, Zkne, Zknd, Zknh, Zkn, Zkt, Zkr. In presence of Zkn the subsets must be ignored in the ISA string.")
        err = True
    return (extension_list, err, err_list)


        



from cerberus import Validator
from riscv_config.warl import warl_interpreter
import riscv_config.constants as constants
from riscv_config.isa_validator import *
import re


class schemaValidator(Validator):
    ''' Custom validator for schema having the custom rules necessary for implementation and checks.'''

    def __init__(self, *args, **kwargs):
        global rv32
        global rv64
        global extensions
        global xlen
        global supported_xlen
        supported_xlen = kwargs.get('xlen')
        xlen = 0 if len(supported_xlen)==0 else max(supported_xlen)
        global isa_string
        isa_string = kwargs.get('isa_string')
        if 32 in supported_xlen:
            rv32 = True
        else:
            rv32 = False
        if 64 in supported_xlen:
            rv64 = True
        else:
            rv64 = False
        super(schemaValidator, self).__init__(*args, **kwargs)

    def _check_with_isa_xlen(self, field, value):
        global supported_xlen
        global isa_string
        if str(max(supported_xlen)) not in isa_string:
            self._error(field, 'XLEN in ISA and supported_xlen fields do not match')

    def _check_with_phy_addr(self, field, value):
        if rv32 and value > 34:
            self._error(field, "Physical address size should not exceed 34 for RV32")
        if rv64 and value > 56:
            self._error(field, "Physical address size should not exceed 56 for RV64")

    def _check_with_cache_block_size(self, field, value):
        if value & (value - 1) != 0:
            self._error(field, "Cache block size should be power of 2")

    def _check_with_cannot_be_false_rv64(self, field, value):
        ''' Functions ensures that the field cannot be False in rv64 mode'''
        if rv64 and not value:
            self._error(field, "This field cannot be False")

    def _check_with_cannot_be_false_rv32(self, field, value):
        ''' Functions ensures that the field cannot be False in rv32 mode'''
        if rv32 and not value:
            self._error(field, "This field cannot be False")

    def _check_with_capture_isa_specifics(self, field, value):
        '''
        Function to extract and store ISA specific information(such as xlen,user
        spec version and extensions present)
        and check whether the dependencies in ISA extensions are satisfied.
        '''
        global xlen
        global extensions
        extension_enc = list("00000000000000000000000000")
        if "32" in value:
            xlen = 32
            ext = value[4:]
        elif "64" in value:
            xlen = 64
            ext = value[4:]
        elif "128" in value:
            xlen = 128
            ext = value[5:]
        else:
            self._error(field, "Invalid width in ISA.")

        if not constants.isa_regex.match(value):
            self._error(field, 'Input ISA string does not match regex')

        (extension_list, err, err_list) = get_extension_list(value)
        if err:
            for e in err_list:
                self._error(field, e)

        #ISA encoding for future use.
        for x in "ABCDEFHIJKLMNPQSTUVX":
            if (x in ext):
                extension_enc[25 - int(ord(x) - ord('A'))] = "1"
        extensions = int("".join(extension_enc), 2)
        extensions = int("".join(extension_enc), 2)

    def _check_with_rv32_check(self, field, value):
        global xlen
        if value:
            if not rv32:
                self._error(
                    field,
                    "Register cannot be implemented in rv32 mode due to unsupported xlen."
                )

    def _check_with_rv64_check(self, field, value):
        global xlen
        if value:
            if not rv64:
                self._error(
                    field,
                    "Register cannot be implemented in rv64 mode due to unsupported xlen."
                )

    def _check_with_max_length(self, field, value):
        '''Function to check whether the given value is less than the maximum value that can be stored(2^xlen-1).'''
        global supported_xlen
        global extensions
        maxv = max(supported_xlen)
        if value > (2**maxv) - 1:
            self._error(field, "Value exceeds max supported length")
    
    def _check_with_max_length32(self, field, value):
        '''Function to check whether the given value is less than the maximum value that can be stored(2^xlen-1).'''
        maxv = 32
        if value > (2**maxv) - 1:
            self._error(field, "Value exceeds max supported length")

    def _check_with_xtveccheck(self, field, value):
        '''Function to check whether the inputs in range type in mtvec are valid.'''
        global xlen
        maxv = 2**(xlen - 2)
        for list in value:
            if (len(list) > 2):
                self._error(field,
                            "Only two values are allowed in each sub list.")
            for val in list:
                if not (val < maxv):
                    self._error(field, "Invalid values.")

    def _check_with_s_check(self, field, value):
        s = 18
        check = False
        if 'implemented' in value:
            if value['implemented']:
                check = True
        if 'accessible' in value:
            if value['accessible']:
                check = True

        if rv64 and check:
            mxl = format(extensions, '#066b')
            if (mxl[65 - s:66 - s] != '1'):
                self._error(field, "should not be implemented since S is not present")

        elif rv32 and check:
            mxl = format(extensions, '#034b')
            if (mxl[33 - s:34 - s] != '1'):
                self._error(field, "should not be implemented S is not present")
    
    def _check_with_fs_check(self, field, value):
        f = 5
        s = 18
        check = False
        if 'implemented' in value:
            if value['implemented']:
                check = True
        if 'accessible' in value:
            if value['accessible']:
                check = True

        if rv64 and check:
            mxl = format(extensions, '#066b')
            if (mxl[65 - s:66 - s] != '1') and (mxl[65 - f:66 - f] != '1'):
                self._error(field, "neither S nor F is present")

        elif rv32 and check:
            mxl = format(extensions, '#034b')
            if (mxl[33 - s:34 - s] != '1') and (mxl[33 - f:34 - f] != '1'):
                self._error(field, "neither S nor F is present")
                

    def _check_with_u_check(self, field, value):
        u = 20
        check = False
        if 'implemented' in value:
            if value['implemented']:
                check = True
        if 'accessible' in value:
            if value['accessible']:
                check = True
        if rv64 and check:
            mxl = format(extensions, '#066b')
            if (mxl[65 - u:66 - u] != '1'):
                self._error(field, "should not be implemented since U is not present")

        elif rv32 and check:
            mxl = format(extensions, '#034b')
            if (mxl[33 - u:34 - u] != '1'):
                self._error(field, "should not be implemented since U is not present")
              
    def _check_with_s_debug_check(self, field, value):
        ''' Function ensures that the ro_constant is hardwired to zero when S is present in the ISA string
            Used mainly for debug schema'''
        global isa_string

        if 'S' not in isa_string :
          if 'ro_constant' not in value:
              self._error(field, "S is not present to dcsr.v should be ro_constant = 0")
          elif value['ro_constant'] != 0:
                self._error(field, "S is not present but ro constant is not hardwired to zero")
                
    def _check_with_u_debug_check(self, field, value):
        ''' Function ensures that the ro_constant is hardwired to zero when U is present in the ISA string
            Used mainly for debug schema'''
        global isa_string

        if 'U' not in isa_string :
          if value['ro_constant'] != 0:
                self._error(field, "U is not present but ro constant is not hardwired to zero")

    def _check_with_su_check(self, field, value):
        s = 18
        u = 20
        check = False
        if 'implemented' in value:
            if value['implemented']:
                check = True
        if 'accessible' in value:
            if value['accessible']:
                check = True

        if rv64 and check:
            mxl = format(extensions, '#066b')
            if (mxl[65 - s:66 - s] != '1') and (mxl[65 - u:66 - u] != '1'):
                self._error(field, "neither S nor U is present")

        elif rv32 and check:
            mxl = format(extensions, '#034b')
            if (mxl[33 - s:34 - s] != '1') and (mxl[33 - u:34 - u] != '1'):
                self._error(field, "neither S nor U is present")
                
    def _check_with_reset_ext(self, field, value):
        
        if rv64:
            mxl = format(extensions, '#066b')
            reset = format(value, '#066b')
            if (mxl[40:66] != reset[40:66] ):
                self._error(field, "reset value does not match with extensions enabled")

        elif rv32 :
            mxl = format(extensions, '#034b')
            reset = format(value, '#034b')
            if (mxl[8:34] != reset[8:34] ):
                self._error(field, "reset value does not match with extensions enabled")
                
    def _check_with_sn_check(self, field, value):
        s = 18
        n = 13
        check = False
        if 'implemented' in value:
            if value['implemented']:
                check = True
        if 'accessible' in value:
            if value['accessible']:
                check = True

        if rv64 and check:
            mxl = format(extensions, '#066b')
            if (mxl[65 - s:66 - s] != '1') and (mxl[65 - n:66 - n] != '1'):
                self._error(field, "neither S nor N is present")

        elif rv32 and check:
            mxl = format(extensions, '#034b')
            if (mxl[33 - s:34 - s] != '1') and (mxl[33 - n:34 - n] != '1'):
                self._error(field, "neither S nor N is present")

    def _check_with_n_check(self, field, value):
        n = 13
        check = False
        if 'implemented' in value:
            if value['implemented']:
                check = True
        if 'accessible' in value:
            if value['accessible']:
                check = True
        if rv64 and check:
            mxl = format(extensions, '#066b')
            if (mxl[65 - n:66 - n] != '1'):
                self._error(field, "should not be implemented since N is not present")

        elif rv32 and check:
            mxl = format(extensions, '#034b')
            if (mxl[33 - n:34 - n] != '1'):
                self._error(field, "should not be implemented since N is not present")

    def _check_with_h_check(self, field, value):
        h = 7
        check = False
        if 'implemented' in value:
            if value['implemented']:
                check = True
        if 'accessible' in value:
            if value['accessible']:
                check = True
        if rv64 and check:
            mxl = format(extensions, '#066b')
            if (mxl[65 - h:66 - h] != '1'):
                self._error(field, "h is not present")

        elif rv32 and check:
            mxl = format(extensions, '#034b')
            if (mxl[33 - h:34 - h] != '1'):
                self._error(field, "h is not present")

    def _check_with_mdeleg_checks(self, field, value):
        if rv32:
            if (value['rv32']['accessible'] == True and
                (not 'S' in self.document['ISA'] and
                 not 'N' in self.document['ISA'])):
                value['rv32']['accessible'] = False
                self._error(field, "S and N are not present")

        if rv64:
            if (value['rv64']['accessible'] == True and
                (not 'S' in self.document['ISA'] and
                 not 'N' in self.document['ISA'])):
                value['rv64']['accessible'] = False
                self._error(field, "S and N are not present")

    def _check_with_ndeleg_checks(self, field, value):
        if rv32:
            if (value['rv32']['accessible'] == True and
                    not 'N' in self.document['ISA']):
                value['rv32']['accessible'] = False
                self._error(field, "should not be implemented since N is not present")

        if rv64:
            if (value['rv64']['accessible'] == True and
                    not 'N' in self.document['ISA']):
                value['rv64']['accessible'] = False
                self._error(field, "should not be implemented since N is not present")

    def _check_with_xcause_check(self, field, value):
        '''Function to verify the inputs for mcause.'''
        if (min(value) < 16):
            self._error(
                field, "Invalid platform specific values for exception cause.")

    def _check_with_wr_illegal32(self, field, value):
        '''Function to ensure the warl does not cross 2^32
        '''
        if 'warl' in value:
            warlnode = warl_interpreter(value['warl'])
            x = 0x8000000000000000
            if (len(value['warl']['legal']) > 1):
                while (x > 0x10000000):
                    if warlnode.islegal(x):
                        self._error(
                            field, "Must have a warl which is only 32-bits. Value"  + \
                                        str(hex(x)) + " was considered legal")
                        break
                    x = x>> 1
            else:
                for l in value['warl']['legal']:
                    if 'bitmask' in l:
                        bmask = re.findall(r'\s*\[.*\]\s*bitmask\s*\[(.*?)\]',l)[0]
                        maxval_str = bmask.split(',')[0]
                        if '0x' in maxval_str:
                            maxval = int(maxval_str,16)
                        else:
                            maxval = int(maxval_str,10)
                        if maxval > 0xFFFFFFFF:
                            self._error(
                                field, "Must have a warl which is only 32-bits. Value"  + \
                                        str(hex(maxval)) + " was considered legal")
                    else:
                        while (x > 0x10000000):
                            if warlnode.islegal(x):
                                self._error(
                                    field, "Must have a warl which is only 32-bits. Value"  + \
                                                str(hex(x)) + " was considered legal")
                                break
                            x = x>> 1

    def _check_with_wr_illegal(self, field, value):
        pr = 0
        pri = 0
        if 'warl' not in value.keys():
            return
        for i in range(len(value['warl']['legal'])):
            if ' -> ' in value['warl']['legal'][i]:
                pr = 1
                break
        if value['warl']['wr_illegal'] != None:
            for i in range(len(value['warl']['wr_illegal'])):
                split = re.findall(r'^\s*\[(\d)]\s*',
                                   value['warl']['wr_illegal'][i])
                if split != []:
                    pri = 1
                    break
        if value['warl']['dependency_fields'] != []:
            l = (len(value['warl']['legal']))
            f = 0
            for i in range(l):
                if "bitmask" in value['warl']['legal'][i]:
                    f = 1
                    splits = re.findall(
                        r'(\[\d\])\s*->\s*.*\s*\[.*\]\s*{}\s*\[.*?[,|:].*?]'.
                        format("bitmask"), value['warl']['legal'][i])
                    if value['warl']['wr_illegal'] != None:
                        for j in range(len(value['warl']['wr_illegal'])):
                            if splits[0] in value['warl']['wr_illegal'][j]:
                                self._error(
                                    field,
                                    "illegal value does not exist for the given mode{}(bitmask)"
                                    .format(splits[0]))
            if f == 0:
                pass

        elif value['warl']['dependency_fields'] == [] and pr == 1:
            self._error(field, "since dependency_fields is empty no '->' in legal fields")
        elif value['warl']['dependency_fields'] == [] and len(value['warl']['legal']) != 1:
            self._error(field, "There should be only one legal string")
        elif value['warl']['dependency_fields'] == [] and pri == 1:
            self._error(field, "since dependency_fields is empty illegal fields must be defined for each")
        else:
            for l in value['warl']['legal']:
                if 'bitmask' in l:
                    bmask = re.findall(r'\s*\[.*\]\s*bitmask\s*\[(.*?)\]',l)[0]
                    if ',' not in bmask:
                        self._error(field, 'Legal string "'+l+'" has wrong bitmask syntax')
                    if len(bmask.split(','))!=2:
                        self._error(field, 'Legal string "'+l+'" has wrong bitmkask syntax')
                    for v in bmask.split(','):
                        if '0x' not in v:
                            try:
                                isinstance(int(v,10),int)
                            except:
                                self._error(field, 'Value ' +str(v) + ' in Legal string "'+l+'" is not a valid number')


#        elif value['warl']['dependency_fields'] == [] and len(
#                value['warl']['legal']
#        ) == 1 and value['warl']['wr_illegal'] != None and "bitmask" in value[
#                'warl']['legal'][0]:
#            self._error(field, "illegal value cannot exist for bitmask type")

    def _check_with_key_check(self, field, value):
        if value['base']['type']['warl']['dependency_fields'] != []:
            par = re.split(
                "::", value['base']['type']['warl']['dependency_fields'][0])
            if not par[1] in value:
                self._error(field, " {} not present".format(par[1]))

    def _check_with_medeleg_reset(self, field, value):
        global supported_xlen
        s = format(value, '#{}b'.format(supported_xlen[0] + 2))
        if (s[-11:-10]) != '0' and value >= int("0x400", 16):
            self._error(field, " 11th bit must be hardwired to 0")

    def _check_with_sedeleg_reset(self, field, value):
        global supported_xlen
        s = format(value, '#{}b'.format(supported_xlen[0] + 2))
        if (s[-11:-8]) != '000' and value >= int("400", 16):
            self._error(field, " 11,10,9 bits should be hardwired to 0")



import logging
import argparse
import operator

import ruamel
from ruamel.yaml import YAML
import yaml as pyyaml

def hexint_presenter(dumper, data):
    return dumper.represent_int(hex(data))

class NoAliasDumper(pyyaml.Dumper):
    def ignore_aliases(self, data):
        return True


def dump_yaml(foo, outfile, no_anchors=False):
    if no_anchors:
        pyyaml.add_representer(int, hexint_presenter)
        pyyaml.dump(foo, outfile, Dumper=NoAliasDumper)
    else:
        yaml = YAML(typ="rt")
        yaml.default_flow_style = False
        yaml.allow_unicode = True
        yaml.compact(seq_seq=False, seq_map=False)
        yaml.indent = 4
        yaml.block_seq_indent = 2
        yaml.dump(foo, outfile)


def load_yaml(foo, no_anchors=False):

    if no_anchors:
        yaml = YAML(typ="safe")
    else:
        yaml = YAML(typ="rt")
    yaml.default_flow_style = False
    yaml.allow_unicode = True
    yaml.compact(seq_seq=False, seq_map=False)
    yaml.indent = 4
    yaml.block_seq_indent = 2

    try:
        with open(foo, "r") as file:
            return yaml.load(file)
    except ruamel.yaml.constructor.DuplicateKeyError as msg:
        logger = logging.getLogger(__name__)
        error = "\n".join(str(msg).split("\n")[2:-7])
        logger.error(error)
        raise SystemExit


class ColoredFormatter(logging.Formatter):
    """
        Class to create a log output which is colored based on level.
    """

    def __init__(self, *args, **kwargs):
        super(ColoredFormatter, self).__init__(*args, **kwargs)
        self.colors = {
            'DEBUG': '\033[94m',
            'INFO': '\033[92m',
            'WARNING': '\033[93m',
            'ERROR': '\033[91m',
        }

        self.reset = '\033[0m'

    def format(self, record):
        msg = str(record.msg)
        level_name = str(record.levelname)
        name = str(record.name)
        color_prefix = self.colors[level_name]
        return '{0}{1:<9s} {4}: {2}{3}'.format(color_prefix,
                                            '[' + level_name + ']', msg,
                                            self.reset, name)


def setup_logging(log_level):
    """Setup logging

        Verbosity decided on user input

        :param log_level: User defined log level

        :type log_level: str
    """
    numeric_level = getattr(logging, log_level.upper(), None)

    if not isinstance(numeric_level, int):
        print(
            "\033[91mInvalid log level passed. Please select from debug | info | warning | error\033[0m"
        )
        raise ValueError("{}-Invalid log level.".format(log_level))

    logging.basicConfig(level=numeric_level)


class SortingHelpFormatter(argparse.HelpFormatter):

    def add_arguments(self, actions):
        actions = sorted(actions, key=operator.attrgetter('option_strings'))
        super(SortingHelpFormatter, self).add_arguments(actions)


def riscv_config_cmdline_args():
    parser = argparse.ArgumentParser(
        formatter_class=SortingHelpFormatter,
        prog="riscv_config",
        description="RISC-V Configuration Validator")
    parser.add_argument('--version',
                        '-v',
                        help='Print version of RISCV-CONFIG being used',
                        action='store_true')
    parser.add_argument('--isa_spec',
                        '-ispec',
                        type=str,
                        metavar='YAML',
                        default=None,
                        help='The YAML which contains the ISA specs.')
    parser.add_argument('--platform_spec',
                        '-pspec',
                        type=str,
                        metavar='YAML',
                        default=None,
                        help='The YAML which contains the Platfrorm specs.')
    parser.add_argument('--debug_spec',
                        '-dspec',
                        type=str,
                        metavar='YAML',
                        default=None,
                        help='The YAML which contains the debug csr specs.') 
    parser.add_argument('--custom_spec',
                        '-cspec',
                        type=str,
                        metavar='YAML',
                        default=None,
                        help='The YAML which contains the custom csr specs.')                        
    parser.add_argument(
        '--work_dir',
        type=str,
        default="riscv_config_work",
        metavar='DIR',
        help='The name of the work dir to dump the output files to.')
    parser.add_argument('--verbose',
                        action='store',
                        default='info',
                        help='debug | info | warning | error',
                        metavar="")
    parser.add_argument('--no_anchors',
                        action='store_true',
                        help='Unroll/Disable all anchors')
    return parser

def pretty_print_yaml(yaml):
    res = ''''''
    for line in ruamel.yaml.round_trip_dump(yaml, indent=5, block_seq_indent=3).splitlines(True):
        res += line
    return res




import os
import re

root = os.path.abspath(os.path.dirname(__file__))

isa_schema = os.path.join(root, 'schemas/schema_isa.yaml')
debug_schema = os.path.join(root, 'schemas/schema_debug.yaml')
platform_schema = os.path.join(root, 'schemas/schema_platform.yaml')
isa_regex = \
        re.compile("^RV(32|64|128)[IE]+[ACDEFGHJLMNPQSTUVX]*(Zicbom|Zicbop|Zicboz|Zicsr|Zifencei|Zihintpause|Zam|Ztso|Zba|Zbb|Zbc|Zbe|Zbf|Zbkb|Zbkc|Zbkx|Zbm|Zbp|Zbr|Zbs|Zbt|Zk|Zkn|Zknd|Zkne|Zknh|Zkr|Zks|Zksed|Zksh|Zkt|Zmmul|Svnapot){,1}(_Zicbom){,1}(_Zicbop){,1}(_Zicboz){,1}(_Zicsr){,1}(_Zifencei){,1}(_Zihintpause){,1}(_Zmmul){,1}(_Zam){,1}(_Zba){,1}(_Zbb){,1}(_Zbc){,1}(_Zbe){,1}(_Zbf){,1}(_Zbkb){,1}(_Zbkc){,1}(_Zbkx){,1}(_Zbm){,1}(_Zbp){,1}(_Zbr){,1}(_Zbs){,1}(_Zbt){,1}(_Zk){,1}(_Zkn){,1}(_Zknd){,1}(_Zkne){,1}(_Zknh){,1}(_Zkr){,1}(_Zks){,1}(_Zksed){,1}(_Zksh){,1}(_Zkt){,1}(_Ztso){,1}(_Svnapot){,1}$")



import logging
import os
import sys
import shutil
from riscv_config import __version__ as version
import riscv_config.checker as checker
import riscv_config.utils as utils
from riscv_config.errors import ValidationError

def main():
    '''
        Entry point for riscv_config.
    '''
    # Set up the parser
    parser = utils.riscv_config_cmdline_args()
    args = parser.parse_args()
    if len(sys.argv) < 2:
        parser.print_help()
        raise SystemExit
    if (args.version):
        print('RISCV-CONFIG: RISC-V Configuration Validator')
        print('Version: ' + version)
        return 0

    # Set up the logger
    utils.setup_logging(args.verbose)
    logger = logging.getLogger()
    logger.handlers = []
    ch = logging.StreamHandler()
    ch.setFormatter(utils.ColoredFormatter())
    logger.addHandler(ch)
    fh = logging.FileHandler('run.log', 'w')
    logger.addHandler(fh)

    work_dir = os.path.join(os.getcwd(), args.work_dir)
    if not os.path.exists(work_dir):
        logger.debug('Creating new work directory: ' + work_dir)
        os.mkdir(work_dir)

    try:
        if args.isa_spec is not None:
            isa_file= checker.check_isa_specs(os.path.abspath(args.isa_spec),
                            work_dir, 
                            True, args.no_anchors)
        if args.platform_spec is not None:
            checker.check_platform_specs(os.path.abspath(args.platform_spec),
                            work_dir, 
                            True, args.no_anchors)
        if args.custom_spec is not None:
            checker.check_custom_specs(os.path.abspath(args.custom_spec),
                            work_dir, 
                            True, args.no_anchors)
        if args.debug_spec is not None:
            if args.isa_spec is None:
             logger.error(' Isa spec missing, Compulsory for debug')
            checker.check_debug_specs(os.path.abspath(args.debug_spec), isa_file,
                            work_dir, 
                            True, args.no_anchors)
    except ValidationError as msg:
        logger.error(str(msg))
        return 1


if __name__ == "__main__":
    exit(main())



import re
import os
import logging
import riscv_config.utils as utils
import textwrap

logger = logging.getLogger(__name__)

class warl_interpreter():
    global val
    global bitsum
    global dep_val

    def __init__(self, warl):
        ''' the warl_description in the yaml is given as input to the constructor '''
        self.warl = warl
        self.dependencies = warl['dependency_fields']
        self.exp  = re.compile('(?P<csr>.*?)\[(?P<csr_ind>.*?)\]\s*(?P<csr_op>.*?)\s*\[(?P<csr_vals>.*?)\]')
        self.rexp = re.compile('(?P<csr>.*?)\[(?P<csr_ind>.*?)\]\s*(in|bitmask)\s*\s*\[(?P<csr_vals>.*?)\]')
        self.rexp_dependencies = re.compile('(?P<dep_csr>.*?)\[(?P<dep_ind>.*?)\]\s*(in|bitmask)\s*\[(?P<dep_vals>.*?)\]\s*->\s*(?P<csr>.*?)\[(?P<csr_ind>.*?)\]\s*(in|bitmask)\s*\s*\[(?P<csr_vals>.*?)\]')

    def prevsum(self, i, k):
        sump = 0
        for j in range(i + 1):
            sump = sump + k[j]
        return sump

    def islegal(self, value, dependency_vals=[]):
        is_legal = False
        logger.debug('Checking for isLegal for WARL: \n' +
                textwrap.indent(utils.pretty_print_yaml(self.warl),'           ')
                        + '\n With following args:'
                        + 'val : ' + str(value) + ' dep_vals :'+
                        str(dependency_vals))
        if not self.dependencies: # no dependencies in the warl
            for legal_str in self.warl['legal']: # iterate over every legal str
                search = self.exp.findall(legal_str)
                if search is None:
                    logger.error('Warl legal string is wrong:' + legal_str)
                    raise SystemExit
                part_legal = False
                for part in search:
                    (csr, csr_ind, csr_op, csr_vals) = part
                    msb = int(csr_ind.split(':')[0])
                    lsb = int(csr_ind.split(':')[-1])
                    bitmask = True if 'bitmask' in csr_op else False

                    if not bitmask: # if its not a bitmask
                        if ":" in csr_vals: # range is specified
                            [base, bound] = csr_vals.split(':')
                            base = int(base, 0)
                            bound = int(bound, 0)
                            if value >= base and value <= bound: # check legal range
                                part_legal = True
                        else:
                            l_vals = csr_vals.split(',')
                            legal_vals = []
                            for i in l_vals : 
                                legal_vals.append(int(i,0))
                            if value in legal_vals:
                                part_legal = True
                    else: # in case of bitmask there are no illegal values
                        part_legal = True
                    if not part_legal:
                        break
                if part_legal:
                    is_legal = True
        else: # there are dependencies
            if len(self.dependencies) != len(dependency_vals):
                logger.error('length of dependency_vals not the same as dependencies')
                raise SystemExit
            leaf_dependencies = []
            for d in self.dependencies:
                leaf_dependencies.append(d.split('::')[-1])
            for legal_str in self.warl['legal']: # iterate over legal str
                dep_str = legal_str.split('->')[0]
                csr_str = legal_str.split('->')[1]
                csr_search = self.exp.findall(csr_str)
                dep_search = self.exp.search(dep_str)
                if csr_search is None or dep_search  is None:
                    logger.error('Warl legal string is wrong :\n\t' + legal_str)
                    raise SystemExit(1)
                dep_csr = dep_search.group('csr')
                dep_ind = dep_search.group('csr_ind')
                if ':' in dep_ind:
                    msb = int(dep_ind.split(':')[0],0)
                    lsb = int(dep_ind.split(':')[1],0)
                else:
                    lsb = int(dep_ind,0)
                    msb = lsb
                if (msb < lsb):
                    logger.error('msb < lsb for in dependency field of warl:\n\n' + 
                            utils.pretty_print_yaml(self.warl))
                    raise SystemExit(1)
                dep_vals = dep_search.group('csr_vals')
                dep_bitmask = True if 'bitmask' in legal_str.split('->')[0] else False

                if dep_csr not in leaf_dependencies:
                    logger.error(\
            'Found a dependency in the legal_str which is not present in the dependency_fields of the warl:\n\n' + \
                    utils.pretty_print_yaml(self.warl))
                    raise SystemExit
                dep_satified = False
                recvd_val = dependency_vals[leaf_dependencies.index(dep_csr)]
                recvd_val = (recvd_val >> lsb) & ((1<<(msb-lsb))-1)

                # check if the dependency value is satisfied.
                if ":" in dep_vals: # range is specified
                    [base, bound] = dep_vals.split(':')
                    base = int(base, 0)
                    bound = int(bound, 0)
                    if recvd_val >= base and recvd_val <= bound: # check legal range
                        dep_satified = True
                else:
                    l_vals = dep_vals.split(',')
                    legal_vals = []
                    for i in l_vals : 
                        legal_vals.append(int(i,0))
                    if recvd_val in legal_vals:
                        dep_satified = True
                
                part_legal = False
                for part in csr_search:
                    (csr, csr_ind, csr_op, csr_vals) = part
                    msb = int(csr_ind.split(':')[0])
                    lsb = int(csr_ind.split(':')[-1])
                    bitmask = True if 'bitmask' in csr_op else False
                    trunc_val = int(bin(value)[2:].zfill(64)[::-1][lsb:msb+1][::-1],base=2)

                    if not bitmask: # if its not a bitmask
                        if ":" in csr_vals: # range is specified
                            [base, bound] = csr_vals.split(':')
                            base = int(base, 0)
                            bound = int(bound, 0)
                            if trunc_val >= base and trunc_val <= bound: # check legal range
                                part_legal = True
                        else:
                            l_vals = csr_vals.split(',')
                            legal_vals = []
                            for i in l_vals : 
                                legal_vals.append(int(i,0))
                            if trunc_val in legal_vals:
                                part_legal = True
                    else: # in case of bitmask there are no illegal values
                        part_legal = True
                    if not part_legal:
                        break
                if part_legal and dep_satified:
                    is_legal = True
                
        return is_legal

    def update(self, curr_val, wr_val, dependency_vals=[]):
        ''' The function takes the current value, write value and an optional list(optional incase there are no dependencies) containing the
        values of the corresponding dependency fields and models the updation of
        the register i.e if the supplied value is legal then the value is returned, else the new value of
        the register is calculated and returned.
        '''
        flag1 = 0
        flag2 = 0
        flag3 = 0
        flag4 = 0
        j = 0
        if self.dependencies != [] and dependency_vals != []:
            for i in range(len(self.warl['legal'])):
                mode1 = re.findall('\[(\d)\]\s*->\s*', self.warl['legal'][i])
                for i1 in range(len(dependency_vals)):
                    if dependency_vals[i1] == int(mode1[i1]):
                        j = i
                        flag1 = 1
                        break
            if flag1 == 0:
                print("Dependency vals do not match")
                exit()
        else:
            if len(self.warl['legal']) != 1:
                print("There cannot be more than one legal value")
                exit()
            else:
                j = 0
        inp1 = self.warl['legal'][j]
        flag1 = 0
        if self.dependencies != [] and dependency_vals != []:
            for i in range(len(self.warl['legal'])):
                if "bitmask" in self.warl['legal'][i]:
                    flag4 = 1
                else:
                    flag4 = 0
                    break

            if flag4 == 0 and not "bitmask" in inp1:
                for i in range(len(self.warl['wr_illegal'])):
                    mode = re.findall('\[(\d)\]\s*.*->\s*',
                                      self.warl['wr_illegal'][i])
                    op = re.findall(r'in\s*\[(.*?)\]',
                                    self.warl['wr_illegal'][i])
                    if op != []:
                        z = re.split("\:", op[0])
                    for i1 in range(len(dependency_vals)):
                        if dependency_vals[i1] == int(mode[i1]):
                            if op != []:
                                if int(wr_val, 0) in range(int(z[0], 0),
                                                    int(z[1], 0)):
                                    j = i
                                    flag1 = 1
                                    break
                            else:
                                j = i
                                flag1 = 1
                                break
                    if flag1 == 1:
                        break
                inp = self.warl['wr_illegal'][j]
            else:
                inp = "no value"
                flag1 = 1

        elif self.dependencies == [] and dependency_vals == [] and not "bitmask" in self.warl[
                'legal'][0]:
            for i in range(len(self.warl['wr_illegal'])):
                op = re.findall(r'\s*wr_val\s*in\s*\[(.*?)\]',
                                self.warl['wr_illegal'][i])
                if op != []:
                    z = re.split("\:", op[0])
                    if int(wr_val, 0) in range(int(z[0], 0), int(z[1], 0)):
                        j = i
                        flag1 = 1
                        break

                else:
                    j = i
                    flag1 = 1
                    break

            inp = self.warl['wr_illegal'][j]

        elif self.dependencies != [] and dependency_vals != [] and len(
                self.warl['legal']) == 1 and "bitmask" in self.warl['legal'][0]:
            flag1 = 1

        if flag1 == 0 and dependency_vals != []:
            print("Dependency vals do not match")
            exit()
        if (self.islegal(curr_val, dependency_vals) == False):
            return "Current value must be legal"
        if (self.islegal(wr_val, dependency_vals)):
            return wr_val
        elif not "bitmask" in inp1:
            if "->" in inp:
                op2 = re.split(r'\->', inp)
                wr = op2[1]
            else:
                wr = inp.strip()
            if "0x" in wr:
                return wr.strip()
            elif wr.lower().strip() == "unchanged":
                return ("0x" + curr_val)

            elif wr.lower().strip() == "nearup":
                a = []
                flag2 = 0
                l = self.legal(dependency_vals)
                for i in range(len(l)):
                    if len(l[i]) == 1:
                        a.append(abs(int(wr_val, 0) - int(l[i][0], 0)))
                for i in range(len(a) - 1, -1, -1):
                    if a[i] == min(a):
                        j = i
                        flag2 = 1
                        break
                if flag2 == 1:
                    return l[j][0]

            elif wr.lower().strip() == "neardown":
                a = []
                l = self.legal(dependency_vals)
                for i in range(len(l)):
                    if len(l[i]) == 1:
                        a.append(abs(int(wr_val, 0) - int(l[i][0], 0)))
                for i in range(len(a)):
                    if a[i] == min(a):
                        j = i
                        flag2 = 1
                        break
                if flag2 == 1:
                    return l[j][0]

            elif wr.lower().strip() == "nextup":
                l = self.legal(dependency_vals)
                for i in range(len(l)):
                    if int(l[i][0], 0) > int(wr_val, 0) and len(l[i]) == 1:
                        j = i
                        flag2 = 1
                        break
                if flag2 == 1:
                    return l[j][0]
                else:
                    return max(l)

            elif wr.lower().strip() == "nextdown":
                l = self.legal(dependency_vals)
                for i in range(len(l)):
                    if int(l[i][0], 0) > int(wr_val, 0) and len(l[i]) == 1:
                        j = i
                        flag2 = 1
                        break
                if flag2 == 1 and j != 0:
                    return l[j - 1][0]
                else:
                    return min(l)

            elif wr.lower().strip() == "max":
                flag3 = 0
                l = self.legal(dependency_vals)
                for i in range(len(l)):
                    if "," in l[i][0]:
                        flag3 = 1
                        j = i
                    else:
                        flag3 = 0
                if flag3 == 0:
                    return max(l)
                else:
                    y = re.split(",", l[j][0])
                    return y[1]

            elif wr.lower().strip() == "min":
                flag3 = 0
                l = self.legal(dependency_vals)
                for i in range(len(l)):
                    if "," in l[i][0]:
                        flag3 = 1
                        j = i
                    else:
                        flag3 = 0
                if flag3 == 0:
                    return min(l)
                else:
                    y = re.split(",", l[j][0])
                    return y[0]

            elif wr.lower().strip() == "addr":
                wr = format(int(wr_val, 0),
                            '#0{}b'.format(4 * self.bitsum + 2))
                wr = wr[2:]
                if wr[0:1] == '0':
                    wr_final = '1' + wr[1:]
                elif wr[0:1] == '1':
                    wr_final = '0' + wr[1:]
                else:
                    print("Invalid binary bit")
                return hex(int(wr_final, 0))

            else:
                return "Invalid update mode"

        else:
            x = re.findall(
                r'\s*.*\s*\[.*\]\s*{}\s*\[(.*?,.*?)\]'.format("bitmask"), inp1)
            z = re.findall(
                r'\s*.*\s*\[(.*)\]\s*{}\s*\[.*?,.*?\]'.format("bitmask"), inp1)
            y = re.split("\,", x[0])
            bitmask = int(y[0], 0)
            fixedval = int(y[1], 0)
            currval = int(wr_val, 0)
            legal = ((currval & bitmask) | fixedval)
            return hex(legal)

    def legal(self, dependency_vals=[]):
        '''The function takes a range(defined as a 2 tuple list) and an optional(optional incase there are no dependencies) list containing the
        values of the corresponding dependency fields and returns the set of legal values as a list of two tuple lists.
        '''
        flag1 = 0
        j = 0
        if self.dependencies != [] and dependency_vals != []:
            for i in range(len(self.warl['legal'])):
                mode = re.findall('\[(\d)\]\s*->\s*', self.warl['legal'][i])
                for i1 in range(len(dependency_vals)):
                    if dependency_vals[i1] == int(mode[i1]):
                        j = i
                        flag1 = 1
                        break
                if flag1 == 1:
                    break
            if flag1 == 0 and dependency_vals != []:
                print("Dependency vals do not match")
                exit()
        else:
            if len(self.warl['legal']) != 1:
                print("There cannot be more than one legal value")
                exit()
            else:
                j = 0
        inp = self.warl['legal'][j]
        s = re.findall(r'in\s*\[(.*?)\]', inp)
        a = []
        b = []
        tup = []
        for i in range(len(s)):
            tup = []
            if ":" in s[i]:
                tup.append(s[i].replace(":", ","))
                a.append(tup)

            else:
                tup = []
                tup.append(s[i])
                a.append(tup)
        if not "bitmask" in inp:
            w = re.split(",", a[0][0])
            o = []
            for i in range(len(w)):
                e = w[i].strip().split()
                o.append(e)
            o.sort()
            return o
        else:
            return a
#str1=\
#'''
#dependency_fields: []
#legal:
#  - "extensions[7:4] in [1,3,5] extensions[3:0] in [0x0:0xF]"
#wr_illegal:
#  - "Unchanged"
#'''
#
#str2=\
#'''
#dependency_fields: []
#legal:
#  - "asid[8:0] in [0x00:0x1FF]"
#wr_illegal:
#  - Unchanged
#'''
#
#node = yaml.load(str2)
#warl = warl_interpreter(node)
#print(str(warl.islegal(0xFFFF, [])))

